# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import textwrap
from qtpy import QtCore, QtWidgets
from libopensesame.oslogging import oslogger
from libqtopensesame.widgets.item_view_button import item_view_button
from libqtopensesame.widgets.tree_item_item import tree_item_item
from libqtopensesame.widgets.qtitem_splitter import qtitem_splitter
from libqtopensesame.widgets import header_widget
from libqtopensesame._input.pool_select import pool_select
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'qtitem', category=u'core')


def requires_init(fnc):
    r"""A decorator that makes sure that an item's controls are initialized
    before a function is called.
    """
    def inner(self, *args, **kwargs):

        if self.container_widget is None:
            self.init_edit_widget()
        return fnc(self, *args, **kwargs)

    return inner


def wait_cursor(fnc):
    r"""A decorator that shows a waiting cursor during the function"""
    def inner(self, *args, **kwargs):

        if self.main_window.is_busy():
            return fnc(self, *args, **kwargs)
        self.main_window.set_busy(True)
        retval = fnc(self, *args, **kwargs)
        self.main_window.set_busy(False)
        return retval

    return inner


class QtItem:
    """Base class for the GUI controls of other items"""
    
    initial_view = u'controls'
    label_align = u'right'
    help_url = None
    lazy_init = True

    def __init__(self):
        """Constructor"""
        
        # The auto-widgets are stored in name -> (var, widget) dictionaries
        self.auto_line_edit = {}
        self.auto_combobox = {}
        self.auto_spinbox = {}
        self.auto_slider = {}
        self.auto_editor = {}
        self.auto_checkbox = {}
        # Lazy initialization means that the control widgets are initialized
        # only when they are shown for the first time. This dramatically speeds
        # up opening of files. However, with some items this is not currently
        # possible, because their widgets are also referred when they are not
        # shown. Currently this means the inline_script.
        if self.lazy_init:
            self.container_widget = None
        else:
            self.init_edit_widget()
        self._dirty = False
        self._shown = False
        self.lock = False
        self.maximized = False
        self.set_validator()
        oslogger.debug(u'created %s' % self.name)
        self.cached_script = self.to_string()
        
    @property
    def plugin_attribute(self):
        return self.plugin_manager[self.item_type].attribute
        
    @property
    def main_window(self):
        return self.experiment.main_window

    @property
    def overview_area(self):
        return self.experiment.main_window.ui.itemtree

    @property
    def theme(self):
        return self.experiment.main_window.theme

    @property
    def tabwidget(self):
        return self.experiment.main_window.tabwidget

    @property
    def console(self):
        return self.experiment.main_window.console

    @property
    def extension_manager(self):
        return self.experiment.main_window.extension_manager

    @property
    def default_description(self):
        return _(u'Default description')

    def open_tab(self, select_in_tree=True, **kwargs):
        """Opens the tab if it wasn't yet open, and switches to it."""
        unsaved = self.main_window.unsaved_changes
        self.tabwidget.add(self.widget(), self.item_icon(), self.name)
        self.main_window.set_unsaved(unsaved)
        if select_in_tree:
            self.experiment.main_window.ui.itemtree.select_item(
                self.name,
                open_tab=False
            )

    def close_tab(self):
        r"""Closes the tab if it was open."""
        self.tabwidget.remove(self.widget())

    def set_focus(self):
        r"""Gives focus to the most important widget."""
        if hasattr(self, u'focus_widget') and self.focus_widget is not None:
            self.focus_widget.setFocus()

    def set_focus_widget(self, widget, override=False):
        r"""Sets the widget that receives focus when the tab is opened.

        Parameters
        ----------
        widget : QWidget, NoneType
            The widget to receive focus or `None` to reset.
        override : bool, optional
            Indicates whether the focus widget should be changed if there
            already is a focus widget.
        """
        if override or not hasattr(self, u'focus_widget') or \
                self.focus_widget is None:
            self.focus_widget = widget

    def update_item_icon(self):
        r"""Updates the item icon."""
        self.tabwidget.set_icon(self.widget(), self.item_icon())
        self.experiment.items.set_icon(self.name, self.item_icon())
        self.header_item_icon.setPixmap(
            self.theme.qpixmap(self.item_icon(), 32))

    def item_icon(self):
        """
        returns:
                desc:	The name of the item icon.
                type:	unicode
        """
        return self.item_type

    def show_tab(self):
        """Is called when the tab becomes visible, and updated the contents."""
        # The first time that the tab is shown, the view is determined. For
        # the controls view, nothing needs to be done, because this is the
        # default view after initialization.
        if not self._shown:
            self._shown = True
            self.main_window.ui.itemtree.inhibit_drag = True
            oslogger.debug(u'Invalid initial_view: %s' % self.initial_view)
            if self.initial_view == u'script':
                self.set_view_script()
            elif self.initial_view == u'split':
                self.set_view_split()
            elif self.initial_view != u'controls':
                oslogger.warning(u'Invalid initial_view: %s' %
                                 self.initial_view)
        self.extension_manager.fire(u'prepare_open_item', name=self.name)
        self.update_script()
        self.edit_widget()
        self.main_window.ui.itemtree.select_item(self.name, open_tab=False)
        self.extension_manager.fire(u'open_item', name=self.name)

    @requires_init
    def widget(self):
        """
        returns:
                desc:	The widget that is added to the tabwidget.
                type:	QWidget
        """
        return self.container_widget

    def init_edit_widget(self, stretch=True):
        r"""Builds the UI.

        Parameters
        ----------
        stretch : bool, optional
            Indicates whether a vertical stretch should be added to the bottom
            of the controls. This is necessary if the controls don't expand.
        """
        oslogger.debug(u'initializing controls for {}'.format(self.name))
        # Header widget
        self.header = header_widget.header_widget(self)
        self.header_hbox = QtWidgets.QHBoxLayout()
        self.header_item_icon = self.theme.qlabel(self.item_icon())
        self.header_hbox.addWidget(self.header_item_icon)
        self.header_hbox.addWidget(self.header)
        self.header_hbox.setContentsMargins(0, 0, 0, 0)
        self.header_hbox.setSpacing(12)
        # Maximize button
        self.button_toggle_maximize = QtWidgets.QPushButton(
            self.theme.qicon(u'view-fullscreen'), u'')
        self.button_toggle_maximize.setToolTip(_(u'Toggle pop-out'))
        self.button_toggle_maximize.setIconSize(QtCore.QSize(16, 16))
        self.button_toggle_maximize.clicked.connect(self.toggle_maximize)
        self.header_hbox.addWidget(self.button_toggle_maximize)
        # View button
        self.button_view = item_view_button(self)
        self.header_hbox.addWidget(self.button_view)
        # Help button
        self.button_help = QtWidgets.QPushButton(
            self.theme.qicon(u"help"), u"")
        self.button_help.setToolTip(
            _(u"Tell me more about the %s item") % self.item_type)
        self.button_help.setIconSize(QtCore.QSize(16, 16))
        self.button_help.clicked.connect(self.open_help_tab)
        self.header_hbox.addWidget(self.button_help)
        self.header_widget = QtWidgets.QWidget()
        self.header_widget.setLayout(self.header_hbox)
        # The edit_grid is the layout that contains the actual controls for the
        # items.
        self.edit_grid = QtWidgets.QFormLayout()
        if self.label_align == u'right':
            self.edit_grid.setLabelAlignment(QtCore.Qt.AlignRight)
        self.edit_grid.setFieldGrowthPolicy(
            QtWidgets.QFormLayout.FieldsStayAtSizeHint)
        self.edit_grid.setContentsMargins(0, 0, 0, 0)
        self.edit_grid.setVerticalSpacing(6)
        self.edit_grid.setHorizontalSpacing(12)
        self.edit_grid_widget = QtWidgets.QWidget()
        self.edit_grid_widget.setLayout(self.edit_grid)
        # The edit_vbox contains the edit_grid and the header widget
        self.edit_vbox = QtWidgets.QVBoxLayout()
        self.edit_vbox.addWidget(self.edit_grid_widget)
        self.edit_vbox.setContentsMargins(0, 0, 0, 0)
        self.edit_vbox.setSpacing(12)
        if stretch:
            self.edit_vbox.addStretch()
        self._edit_widget = QtWidgets.QWidget()
        self._edit_widget.setWindowIcon(self.theme.qicon(self.item_type))
        self._edit_widget.setLayout(self.edit_vbox)
        # The script widget is not initialized yet, to improve performance.
        # This will happen only when the script is shown.
        self._script_widget = None
        self._script_frame = QtWidgets.QWidget()
        # The container_widget is the top-level widget that is actually
        # inserted into the tab widget.
        self.splitter = qtitem_splitter(self)
        self.set_view_controls()
        self.splitter.splitterMoved.connect(self.splitter_moved)
        self.container_vbox = QtWidgets.QVBoxLayout()
        self.container_vbox.setContentsMargins(12, 12, 12, 12)
        self.container_vbox.setSpacing(18)
        self.container_vbox.addWidget(self.header_widget)
        self.container_vbox.addWidget(self.splitter)
        self.container_widget = QtWidgets.QWidget()
        self.container_widget.setLayout(self.container_vbox)
        self.container_widget.on_activate = self.show_tab
        self.container_widget.__item__ = self.name

    @wait_cursor
    def init_script_widget(self):

        from libqtopensesame.pyqode_extras.widgets import \
            OpenSesameCodeEdit

        oslogger.debug(u'initializing script widget for {}'.format(self.name))
        self._script_widget = OpenSesameCodeEdit(self.main_window)
        self._script_widget.focusOutEvent = self._script_focus_out
        self.extension_manager.fire(
            u'register_editor',
            editor=self._script_widget
        )
        self._script_button = QtWidgets.QPushButton(
            self.theme.qicon(u'dialog-apply'),
            _(u'Apply')
        )
        self._script_button.is_apply_button = True
        self._script_button.clicked.connect(
            self.apply_script_changes_and_switch_view
        )
        self._script_button_layout = QtWidgets.QHBoxLayout()
        self._script_button_layout.setContentsMargins(0, 0, 0, 0)
        self._script_button_layout.addStretch()
        self._script_button_layout.addWidget(self._script_button)
        self._script_button_widget = QtWidgets.QWidget()
        self._script_button_widget.setLayout(self._script_button_layout)
        self._script_layout = QtWidgets.QVBoxLayout()
        self._script_layout.addWidget(self._script_button_widget)
        self._script_layout.addWidget(self._script_widget)
        self._script_layout.setAlignment(QtCore.Qt.AlignRight)
        self._script_layout.setContentsMargins(0, 12, 0, 0)
        self._script_layout.setSpacing(12)
        self._script_frame.setLayout(self._script_layout)
        self.update_script(use_cache=False)

    def _script_focus_out(self, event):
        r"""Is called when the script editor loses focus, so that changes can
        be applied.

        Parameters
        ----------
        event
            a QFocusEvent
        """
        # When the apply button is pressed, the script is also applied. But this
        # also triggers a focusOutEvent. Hence, this check avoids applying the
        # script twice.
        w = self.main_window.focusWidget()
        if (
                not isinstance(w, QtWidgets.QPushButton) or
                not hasattr(w, 'is_apply_button')
        ):
            self.apply_script_changes()

    def splitter_moved(self, pos, index):
        r"""Is called when the splitter handle is manually moved.

        Parameters
        ----------
        pos : int
            The splitter-handle position.
        index : int
            The index of the splitter handle. Since there is only one handle,
            this is always 0.
        """
        sizes = self.splitter.sizes()
        self.edit_size = sizes[0]
        self.script_size = sizes[1]
        if self.script_size == 0:
            self.button_view.set_view_icon(u'controls')
        elif self.edit_size == 0:
            self.button_view.set_view_icon(u'script')
        else:
            self.button_view.set_view_icon(u'split')

    def set_view_controls(self):
        r"""Puts the splitter in control view."""
        self.splitter.setSizes([self.splitter.width(), 0])
        self.button_view.set_view_icon(u'controls')

    def set_view_script(self):
        r"""Puts the splitter in script view."""
        if self._script_widget is None:
            self.init_script_widget()
        self.splitter.setSizes([0, self.splitter.width()])
        self.button_view.set_view_icon(u'script')

    def set_view_split(self):
        r"""Puts the splitter in split view."""
        if self._script_widget is None:
            self.init_script_widget()
        width = self.splitter.width() // 2
        self.splitter.setSizes([width, width])
        self.button_view.set_view_icon(u'split')

    def update(self):
        r"""Updates both the script and the controls."""
        # Items are updated when they are shown in a tab, or as a maximized
        # window, or in a dockwidget. If items are not shown, we don't need to
        # update them, thus improving performance.
        if (
                not self.maximized and
                not self.extension_manager.provide(
                    'item_in_dockwidget',
                    name=self.name
                ) and
                self.tabwidget.current_item() != self.name
        ):
            return False
        self.update_script()
        self.edit_widget()
        return True

    def update_script(self, use_cache=True):
        r"""Regenerates the script and updates the script widget."""
        script = self.to_string()
        if use_cache and self.cached_script == script:
            return  # nothing changed
        self.cached_script = script
        # Normally, the script starts with a 'define' line and is indented by
        # a tab. We want to undo this, and present only unindented content.
        script = script[script.find(u'\t'):]
        script = textwrap.dedent(script)
        self.main_window.set_unsaved()
        if self._script_widget is not None:
            self._script_widget.setPlainText(
                script,
                u'text/generic',
                u'utf-8'
            )
        self.extension_manager.fire(u'change_item', name=self.name)

    def edit_widget(self):
        r"""This function updates the controls based on the item state."""
        self.auto_edit_widget()
        self.header.refresh()

    def apply_edit_changes(self):
        r"""Applies changes to the graphical controls."""
        self.auto_apply_edit_changes()
        self.update_script()
        self.set_clean()
        return True

    def apply_script_changes(self):
        r"""Applies changes to the script."""
        if not self.validate_script():
            self.set_view_script()
            return
        new_script = self._script_widget.toPlainText()
        old_script = self.to_string()
        self.from_string(new_script)
        if old_script != new_script:
            self.main_window.set_unsaved()
            self.cached_script = new_script
        self.edit_widget()

    def apply_script_changes_and_switch_view(self):
        r"""Applies changes to the script if possible. If so, switches to the
        controls view.
        """
        if not self.validate_script():
            self.set_view_script()
            return
        self.apply_script_changes()
        self.set_view_controls()

    def validate_script(self):
        r"""Checks whether the script is syntactically valid. If not, the
        offending line is highlighted if QProgEdit supports setInvalid().

        Returns
        -------
        bool
        """
        script = self._script_widget.toPlainText()
        # First create a dummy item to see if the string can be parsed.
        try:
            self.validator(self.name, self.experiment, script)
            return True
        except Exception as e:
            # If an error occurs, we first parse the first line, then the first
            # and second, and so on, until we find the error.
            l = script.split(u'\n')
            for line_nr, line in enumerate(l):
                test_script = u'\n'.join(l[:line_nr])
                try:
                    self.validator(self.name, self.experiment, test_script)
                except Exception as e_:
                    from pyqode.core.api.utils import TextHelper
                    TextHelper(self._script_widget).select_lines(
                        line_nr - 1,
                        line_nr - 1
                    )
                    break
            self.console.write(e)
            self.extension_manager.fire(
                u'notify',
                message=_(
                    u'Failed to parse script. See debug window for details'),
                category=u'warning',
                always_show=True
            )
            return False

    def set_validator(self):
        r"""Sets the validator class, that is, the class that is used to parse
        scripts to see if they are syntactically correct. Currently, we use the
        class that defines from_string().
        """
        import inspect
        meth = self.from_string
        for cls in inspect.getmro(self.__class__):
            if meth.__name__ in cls.__dict__:
                break
        oslogger.debug(u'validator: %s' % cls)
        self.validator = cls

    def rename(self, from_name, to_name):
        r"""Handles renaming of an item (not necessarily the current item).

        Parameters
        ----------
        from_name : unicode
            The old item name.
        to_name : unicode
            The new item name
        """
        if self.name != from_name:
            return
        self.name = to_name
        if self.container_widget is None:
            return
        self.container_widget.__item__ = self.name
        self.header.set_name(to_name)
        index = self.tabwidget.indexOf(self.widget())
        if index is not None:
            self.tabwidget.setTabText(index, to_name)

    def open_help_tab(self):
        r"""Opens a help tab."""
        if self.help_url is None:
            self.tabwidget.open_help(self.item_type)
        else:
            self.tabwidget.open_osdoc(self.help_url)

    def toggle_maximize(self):
        r"""Toggles edit-widget maximization."""
        if not self.maximized:
            # Always ignore close events. This is necessary, because otherwise
            # the pop-out widget can be closed without re-enabling the main
            # window.
            self.main_window.block_close_event = True
            self.container_widget.closeEvent = lambda e: e.ignore()
            self.container_widget.setParent(None)
            self.container_widget.setWindowFlags(
                QtCore.Qt.WindowStaysOnTopHint |
                QtCore.Qt.WindowMaximizeButtonHint |
                QtCore.Qt.CustomizeWindowHint)
            self.container_widget.showMaximized()
            self.container_widget.show()
            self.button_toggle_maximize.setIcon(
                self.theme.qicon(u'view-restore'))
        else:
            self.main_window.block_close_event = False
            self.container_widget.setParent(self.main_window)
            self.open_tab()
            self.button_toggle_maximize.setIcon(
                self.theme.qicon(u'view-fullscreen'))
        self.maximized = not self.maximized
        self.button_help.setDisabled(self.maximized)
        self.main_window.setDisabled(self.maximized)

    def delete(self, item_name, item_parent=None, index=None):
        r"""Is called when an item is deleted (not necessarily the current
        one).

        Parameters
        ----------
        item_name : str
            The name of the item to be deleted.
        item_parent : str, optional
            The name of the parent item.
        index : int, optional
            The index of the item in the parent.
        """
        pass

    def build_item_tree(
            self,
            toplevel=None,
            items=[],
            max_depth=-1,
            extra_info=None
    ):
        r"""Constructs an item tree.

        Parameters
        ----------
        toplevel : tree_base_item, optional
            The toplevel widget.
        items    A list of item names that have already been added, to prevent
            recursion.

        Returns
        -------
        tree_item_item
        """
        widget = tree_item_item(self, extra_info=extra_info)
        items.append(self.name)
        if toplevel is not None:
            toplevel.addChild(widget)
        return widget

    def parents(self):
        """
        returns:
                desc:	A list of all parents (names) of the current item.
                type:	list
        """
        l = [self.name]
        for item in self.experiment.items:
            if self.experiment.items[item].is_child_item(self.name):
                l.append(item)
        return l

    def get_ready(self):
        r"""This function should be overridden to do any last-minute stuff that
        and item should do before an experiment is actually run, such as
        applying pending script changes.

        Returns
        -------
        bool
            True if some action has been taken, False if nothing was done.
        """
        if self._dirty:
            self.apply_edit_changes()
            return True
        return False

    def auto_edit_widget(self):
        r"""Updates the GUI controls based on the auto-widgets."""
        for var, edit in self.auto_line_edit.items():
            if isinstance(var, int):
                continue
            if var in self.var:
                val = safe_decode(self.var.get(var, _eval=False))
            else:
                val = u''
            if val != edit.text():
                edit.setText(val)

        for var, combobox in self.auto_combobox.items():
            if isinstance(var, int):
                continue
            val = self.var.get(var, _eval=False, default=u'')
            i = combobox.findText(safe_decode(val))
            # Set the combobox to the select item
            if i >= 0:
                if i != combobox.currentIndex() or not combobox.isEnabled():
                    combobox.setDisabled(False)
                    combobox.setCurrentIndex(i)
            # If no value was specified, set the combobox to a blank item
            elif val == u'':
                if combobox.currentIndex() >= 0 or not combobox.isEnabled():
                    combobox.setDisabled(False)
                    combobox.setCurrentIndex(-1)
            # If an unknown value has been specified, notify the user
            else:
                if combobox.isEnabled():
                    combobox.setDisabled(True)
                    self.extension_manager.fire(
                        u'notify',
                        message=_('"%s" is set to a variable or unknown '
                                  'value and can only be edited through '
                                  'the script.') % var, category=u'info')

        for var, spinbox in list(self.auto_spinbox.items()) \
                + list(self.auto_slider.items()):
            if isinstance(var, int):
                continue
            if var in self.var:
                val = self.var.get(var, _eval=False)
                if type(val) in (float, int):
                    if spinbox.value() == val and spinbox.isEnabled():
                        continue
                    spinbox.setDisabled(False)
                    try:
                        spinbox.setValue(val)
                    except Exception as e:
                        self.main_window.notify(
                            _(u"Failed to set control '%s': %s") % (var, e))
                else:
                    if not spinbox.isEnabled():
                        continue
                    spinbox.setDisabled(True)
                    self.extension_manager.fire(
                        u'notify',
                        message=_('"%s" is defined using variables and can '
                                  'only be edited through the script.') % var,
                        category=u'info')

        for var, checkbox in self.auto_checkbox.items():
            if isinstance(var, int):
                continue
            if var in self.var:
                val = self.var.get(var, _eval=False)
                if val in [u'yes', u'no']:
                    checkbox.setDisabled(False)
                    checked = val == u'yes'
                    if checked != checkbox.isChecked():
                        checkbox.setChecked(checked)
                else:
                    checkbox.setDisabled(True)
                    self.extension_manager.fire(
                        u'notify',
                        message=_('"%s" is defined using variables or has an '
                                  'invalid value, and can only be edited '
                                  'through the script.') % var,
                        category=u'info')

        for var, editor in self.auto_editor.items():
            if isinstance(var, int):
                continue
            val = safe_decode(self.var.get(var, _eval=False, default=u''))
            if val != editor.toPlainText():
                editor.setPlainText(val)

    def auto_apply_edit_changes(self):
        r"""Applies the auto-widget controls."""
        for var, edit in self.auto_line_edit.items():
            if isinstance(var, int):
                continue
            if edit.isEnabled() and isinstance(var, str):
                val = edit.text().strip()
                if val:
                    self.var.set(var, val)
                    continue
                # If no text was entered, we use a default if available ...
                if hasattr(edit, u'default'):
                    self.var.set(var, edit.default)
                    continue
                # ... or unset the variable if no default is available.
                self.var.unset(var)

        for var, combobox in self.auto_combobox.items():
            if isinstance(var, int):
                continue
            if combobox.isEnabled() and isinstance(var, str):
                self.var.set(var, combobox.currentText())

        for var, spinbox in self.auto_spinbox.items():
            if isinstance(var, int):
                continue
            if spinbox.isEnabled() and isinstance(var, str):
                self.var.set(var, spinbox.value())

        for var, slider in self.auto_slider.items():
            if isinstance(var, int):
                continue
            if slider.isEnabled() and isinstance(var, str):
                self.var.set(var, slider.value())

        for var, checkbox in self.auto_checkbox.items():
            if isinstance(var, int):
                continue
            if checkbox.isEnabled() and isinstance(var, str):
                val = u'yes' if checkbox.isChecked() else u'no'
                self.var.set(var, val)

        for var, editor in self.auto_editor.items():
            if isinstance(var, int):
                continue
            if isinstance(var, str):
                self.var.set(var, editor.toPlainText())
        return True

    def auto_add_widget(self, widget, var=None, apply_func=None):
        r"""Adds a widget to the list of auto-widgets.

        Parameters
        ----------
        widget : QWidget
        var, optional
            The name of the experimental variable to be linked to the widget,
            or None to use an automatically chosen name.
        type"""
        # Use the object id as a fallback name
        if var is None:
            var = id(widget)
        if apply_func is None:
            apply_func = self.apply_edit_changes
        self.set_focus_widget(widget)
        if (
                isinstance(widget, QtWidgets.QSpinBox) or
                isinstance(widget, QtWidgets.QDoubleSpinBox)
        ):
            widget.editingFinished.connect(apply_func)
            widget.valueChanged.connect(self.set_dirty)
            self.auto_spinbox[var] = widget
        elif isinstance(widget, QtWidgets.QComboBox):
            widget.activated.connect(apply_func)
            self.auto_combobox[var] = widget
        elif (
                isinstance(widget, QtWidgets.QSlider) or
                isinstance(widget, QtWidgets.QDial)
        ):
            widget.valueChanged.connect(apply_func)
            self.auto_slider[var] = widget
        elif (
                isinstance(widget, QtWidgets.QLineEdit) or
                isinstance(widget, pool_select)
        ):
            widget.editingFinished.connect(apply_func)
            widget.textEdited.connect(self.set_dirty)
            self.auto_line_edit[var] = widget
        elif isinstance(widget, QtWidgets.QCheckBox):
            widget.clicked.connect(apply_func)
            self.auto_checkbox[var] = widget
        else:
            raise TypeError(u"Cannot auto-add widget of type %s" % widget)

    def set_dirty(self):
        r"""Indicates that the item's controls have changed since being last
        applied.
        """
        self._dirty = True

    def set_clean(self):
        r"""Indicates that the item's controls have all been applied."""
        self._dirty = False

    def children(self):
        """
        returns:
                desc:	A list of children, including grand children, and so on.
                type:	list
        """
        return []

    def direct_children(self):
        """
        returns:
                desc:	A list of direct children, which does not include grand
                                children, and so on.
                type:	list
        """
        return []

    def is_child_item(self, item_name):
        r"""Checks if an item is somewhere downstream from the current item in
        the experimental hierarchy.

        Parameters
        ----------
        item_name : unicode
            The name of the child item.

        Returns
        -------
        bool
            True if the current item is offspring of the item, False otherwise.
        """
        return False

    def insert_child_item(self, item_name, index=0):
        r"""Inserts a child item, if applicable to the item type.

        Parameters
        ----------
        item_name : unicode
            The name of the child item.
        index : int, optional
            The index of the child item.
        """
        pass

    def remove_child_item(self, item_name, index=0):
        r"""Removes a child item, if applicable to the item type.

        Parameters
        ----------
        item_name : unicode
            The name of the child item.
        index : int, optional
            The index of the child item, if applicable. A negative value
            indicates all instances.
        """
        pass
    
    def enable_child_item(self, item_name, index=0, enabled=True):
        r"""Enables a child item, if applicable to the item type.

        Parameters
        ----------
        item_name : unicode
            The name of the child item.
        index : int, optional
            The index of the child item, if applicable. A negative value
            indicates all instances.
        enabled : bool, optional
        """
        pass


# Alias for backwards compatibility
qtitem = QtItem



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy.QtWidgets import QSizePolicy
from qtpy.QtCore import Qt
import ast
from libopensesame.inline_script import InlineScript as InlineScriptRuntime
from libopensesame.oslogging import oslogger
from libqtopensesame.items.qtplugin import QtPlugin
from libqtopensesame.misc.translate import translation_context
from pyqode.core.api.utils import TextHelper
_ = translation_context(u'inline_script', category=u'item')


class InlineScript(InlineScriptRuntime, QtPlugin):
    """The inline_script GUI controls"""
    
    description = _(u'Executes Python code')
    help_url = u'manual/python/about'
    ext = u'.py'
    mime_type = u'text/x-python'

    def __init__(self, name, experiment, string=None):
        self._var_cache = None
        InlineScriptRuntime.__init__(self, name, experiment, string)
        QtPlugin.__init__(self)

    def apply_edit_changes(self):
        sp = self._pyqode_prepare_editor.toPlainText()
        sr = self._pyqode_run_editor.toPlainText()
        self._set_modified()
        self.var._prepare = sp
        self.var._run = sr
        self._var_cache = None
        super().apply_edit_changes()

    def _set_modified(self, prepare=False, run=False):
        self._pyqode_prepare_editor.document().setModified(prepare)
        self._pyqode_run_editor.document().setModified(run)
        self._pyqode_tab_bar.setTabText(
            0, ('* ' if prepare else '') + _('Prepare'))
        self._pyqode_tab_bar.setTabText(
            1, ('* ' if prepare else '') + _('Run'))

    def set_focus(self):
        self._pyqode_tab_widget.setFocus()

    def init_edit_widget(self):
        from pyqode.core.widgets import SplittableCodeEditTabWidget

        super().init_edit_widget(stretch=False)
        self._pyqode_tab_widget = SplittableCodeEditTabWidget(
            tabs_movable=False,
            plus_button=False,
            tab_context_menu=False,
            empty_context_menu=False
        )
        self._pyqode_tab_widget.setSizePolicy(
            QSizePolicy.Expanding,
            QSizePolicy.Expanding
        )
        self._pyqode_tab_bar = self._pyqode_tab_widget.main_tab_widget.tabBar()
        self._pyqode_tab_bar.setTabsClosable(False)
        self._pyqode_prepare_editor = \
            self._pyqode_tab_widget.create_new_document('Prepare', self.ext)
        self._pyqode_run_editor = \
            self._pyqode_tab_widget.create_new_document('Run', self.ext)
        self._pyqode_run_editor.focusOutEvent = self._editor_focus_out
        self._pyqode_prepare_editor.focusOutEvent = self._editor_focus_out
        self._set_modified()
        self.extension_manager.fire(
            u'register_editor',
            editor=self._pyqode_run_editor
        )
        self.extension_manager.fire(
            u'register_editor',
            editor=self._pyqode_prepare_editor
        )
        self.edit_vbox.addWidget(self._pyqode_tab_widget)
        if not self.var._run and self.var._prepare:
            self._pyqode_tab_widget.main_tab_widget.setCurrentIndex(0)
        else:
            self._pyqode_tab_widget.main_tab_widget.setCurrentIndex(1)

    def edit_widget(self):
        super().edit_widget()
        _prepare = safe_decode(self.var._prepare)
        if _prepare != self._pyqode_prepare_editor.toPlainText():
            self._pyqode_prepare_editor.setPlainText(
                _prepare,
                mime_type=self.mime_type
            )
        _run = safe_decode(self.var._run)
        if _run != self._pyqode_run_editor.toPlainText():
            self._pyqode_run_editor.setPlainText(
                _run,
                mime_type=self.mime_type
            )

    def get_ready(self):
        if self.container_widget is None:
            return
        self.apply_edit_changes()
        
    def open_tab(self, select_in_tree=True, **kwargs):
        super().open_tab(select_in_tree=select_in_tree, **kwargs)
        if 'phase' not in kwargs:
            return
        tab_index = 1 if kwargs['phase'] == 'run' else 0
        self._pyqode_tab_widget.main_tab_widget.setCurrentIndex(tab_index)
        # The line number is always passed as the first optional argument
        if 'args' in kwargs:
            line = int(kwargs['args'][0])
            edit = self._pyqode_tab_widget.main_tab_widget.currentWidget()
            TextHelper(edit).goto_line(line, move=True)

    def var_info(self):
        if self._var_cache is None:
            script = self.var.get('_prepare', _eval=False, default='') + \
                self.var.get('_run', _eval=False, default='')
            self._var_cache = [
                (key, None) for key in self._extract_assignments(script)]
        return super().var_info() + self._var_cache
    
    @staticmethod
    def _extract_assignments(script):
        """Extracts variables that are assigned in the script.
        
        Parameters
        ----------
        script: str
        
        Returns
        -------
        list
            A list of extracted variable names
        """
        def inner(body, only_globals=False):
            assignments = []
            for element in body:
                if isinstance(element, ast.Attribute):
                    assignments += [target.attr for target in element.targets
                                    if element.value == 'var']
                elif isinstance(element, ast.Assign) and not only_globals:
                    for target in element.targets:
                        if isinstance(target, ast.Attribute):
                            # Oldschool assignments to variable store
                            if target.value.id == 'var':
                                assignments.append(target.attr)
                        else:
                            assignments.append(target.id)
                elif isinstance(element, ast.Global):
                    assignments += element.names
                elif hasattr(element, 'body'):
                    assignments += inner(
                        element.body,
                        only_globals=isinstance(element, ast.FunctionDef))
            return assignments
        
        try:
            return inner(ast.parse(script).body)
        except Exception as e:
            oslogger.debug(f'failed to extract assignments: {e}')
            return []


# Alias for backwards compatibility
inline_script = InlineScript



"""Notifies the user if an experiment has been created with a different version
of OpenSesame.
"""

icon = "applications-accessories"
label = "OpenSesame 4 notifications"
priority = -1
settings = {
    "os4n_old_experiment_notification": True
}



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtWidgets
from libqtopensesame.widgets.base_widget import BaseWidget
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'general_script_editor', category=u'core')


class GeneralScriptEditor(BaseWidget):

    r"""The general script editor."""
    def __init__(self, main_window):
        r"""Constructor.

        Parameters
        ----------
        main_window
            A qtopensesame object.
        """
        from libqtopensesame.pyqode_extras.widgets import OpenSesameCodeEdit
        super().__init__(main_window, ui=u'widgets.general_script_editor')
        self.ui.editor = OpenSesameCodeEdit()
        self.extension_manager.fire(
            u'register_editor',
            editor=self.ui.editor,
            mime_type='text/opensesame'
        )
        self.ui.layout_vbox.addWidget(self.ui.editor)
        self.ui.button_apply.clicked.connect(self._apply)
        self.tab_name = u'__general_script__'

    def _apply(self):
        r"""Confirms and applies the script changes."""
        resp = QtWidgets.QMessageBox.question(
            self.main_window,
            _(u'Apply?'),
            _(u'Are you sure you want to apply the changes to the general script?'),
            QtWidgets.QMessageBox.Yes,
            QtWidgets.QMessageBox.No
        )
        if resp == QtWidgets.QMessageBox.No:
            return
        self.main_window.regenerate(self.ui.editor.toPlainText())

    def on_activate(self):
        r"""Refreshes the tab when it is activated."""
        self.refresh()

    def refresh(self):
        r"""Refreshes the contents of the general script."""
        self.ui.editor.setPlainText(
            self.main_window.experiment.to_string(),
            u'text/generic',
            u'utf-8'
        )


# Alias for backwards compatibility
general_script_editor = GeneralScriptEditor



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import textwrap
from libopensesame.widgets._widget import Widget
from openexp.canvas_elements import RichText


class Label(Widget):

    r"""The `Label` widget is a non-interactive string of text.

    __Example
    (OpenSesame script):__

    ~~~
    widget 0 0 1 1 label text='My text'
    ~~~
    __Example (Python):__

    ~~~ .python
    form = Form()
    label = Label(text='My
    text')
    form.set_widget(label, (0,0))
    form._exec()
    ~~~

    [TOC]
    """
    def __init__(self, form, text=u'label', frame=False, center=True):
        r"""Constructor to create a new `Label` object. You do not generally
        call this constructor directly, but use the `Label()` factory
        function,
        which is described here: [/python/common/]().

        Parameters
        ----------
        form : form
            The parent form.
        text : str, unicode, optional
            The label text.
        frame : bool, optional
            Indicates whether a frame should be drawn around the widget.
        center : bool, optional
            Indicates whether the text should be centered.
        """
        if isinstance(frame, str):
            frame = frame == u'yes'
        if isinstance(center, str):
            center = center == u'yes'
        Widget.__init__(self, form)
        self.type = u'label'
        self.text = text
        self.frame = frame
        self.html = True
        self.center = center
        self.x_pad = 8
        self.y_pad = 8
        self.tab_str = u'    '  # Replace tab characters by four spaces

    def _init_canvas_elements(self):

        x, y, w, h = self.rect
        if self.center:
            x += w/2
            y += h/2
        else:
            x += self.x_pad
            y += self.y_pad
        w -= 2*self.x_pad
        self._text_element = RichText(self.text, center=self.center,
                                      x=x, y=y, max_width=w, html=self.html).construct(self.canvas)
        self.canvas.add_element(self._text_element)
        Widget._init_canvas_elements(self)

    def _update_text(self, text):
        r"""Draws text inside the widget.

        Parameters
        ----------
        text : str, unicode
            The text to draw.
        """
        text = self.form.experiment.syntax.eval_text(text)
        text = safe_decode(text).replace(u'\t', self.tab_str)
        self._text_element.text = text

    def _update(self):
        r"""Draws the widget."""
        if self.frame:
            self._update_frame(self.rect)
        self._update_text(self.text)


label = Label



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtWidgets
import os.path


class RecentAction(QtWidgets.QAction):

    """Menu action for a recently opened file"""
    def __init__(self, path, main_window, menu):
        """
        Constructor

        Arguments:
        path -- path to the recent file
        main_window -- the main window
        menu -- the menu into which the action should be inserted
        """
        QtWidgets.QAction.__init__(self, os.path.basename(path), menu)
        self.main_window = main_window
        self.triggered.connect(self.open_file)
        self.path = path

    def open_file(self, dummy=None):
        """
        Open the file

        Keyword arguments:
        dummy -- a dummy argument passed by the signal handler (default=None)
        """
        self.main_window.open_file(path=self.path)


# Alias for backwards compatibility
recent_action = RecentAction



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import os
from openexp import resources
from importlib import import_module
from libopensesame.misc import camel_case
from libopensesame import plugins  # deprecated
from libopensesame.plugin_manager import OldStylePlugin, Plugin, PluginManager
from libopensesame.oslogging import oslogger
from libqtopensesame.items.qtautoplugin import QtAutoPlugin
from libqtopensesame.misc.translate import translation_context


class QtPlugin(Plugin):
    """An extension of the Plugin class that loads the Qt interface of a
    plugin or extension.
    """
    
    def __init__(self, mod):
        super().__init__(mod)
        # The main icon is set to the icon plugin attribute if available, and
        # otherwise to a [plugin_name].png file in the plugin folder
        self.icon = self._mod.icon if hasattr(self._mod, 'icon') \
            else os.path.join(self.folder, f'{self.name}.png')
        # The large icon is set to the large_icon plugin attribute if
        # available, and otherwise to the main icon if specified through an
        # attribute and otherwise to a [plugin_name]_large.png file.
        if hasattr(self._mod, 'icon_large'):
            self.icon_large = self._mod.icon_large
        elif hasattr(self._mod, 'icon'):
            self.icon_large = self._mod.icon
        else:
            self.icon_large = os.path.join(self.folder,
                                           f'{self.name}_large.png')

    @property
    def description(self):
        # We translate the description
        _ = translation_context(self.name, category=self._type)
        return _(super().description)
    
    def _get_cls(self, mod):
        # If there is a class named Qt[PluginName], then is assumed to be
        # the GUI class
        cls_name = f'Qt{camel_case(self.name)}'
        if hasattr(mod, cls_name):
            return getattr(mod, cls_name)
        # Otherwise, we dynamically create a class that inherits from the
        # plugin runtime class and QtAutoPlugin. We also specify a custom
        # __init__() function that calls the parent constructors correctly.
        oslogger.info(
            f'dynamically creating plugin gui class for {self.name}')
        Runtime = getattr(mod, camel_case(self.name))
        
        def cls_init(self, name, experiment, script=None):
            Runtime.__init__(self, name, experiment, script)
            QtAutoPlugin.__init__(self, mod.__file__)
            
        return type(cls_name, (Runtime, QtAutoPlugin), {'__init__': cls_init})
        
    def supports(self, experiment):
        fnc = self.attribute('supports', None)
        if callable(fnc):
            return fnc(experiment)
        return True


class QtOldStylePlugin(OldStylePlugin):
    """An extension of the OldStylePlugin class that loads the Qt interface of a
    plugin or extension. This is deprecated and will be removed.
    """
    
    def __init__(self, name, type_):
        super().__init__(name, type_)
        self.icon = plugins.plugin_icon_small(name, _type=self.type_)
        self.icon_large = plugins.plugin_icon_large(name, _type=self.type_)
    
    def build(self, *args, **kwargs):
        oslogger.debug(f'building old-style plugin gui for {self.name}')
        if self.type_ == 'plugins':
            return plugins.load_plugin(self.name, *args, prefix='qt', **kwargs)
        return plugins.load_extension(self.name, *args, **kwargs)

    def supports(self, experiment):
        return True


class QtPluginManager(PluginManager):
    """An extension of the PluginManager that provides access to the Qt
    interfaces of plugins and extensions.
    """
    
    plugin_cls = QtPlugin
    oldstyle_plugin_cls = QtOldStylePlugin



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtWidgets, QtGui
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'popup_menu', category=u'core')


class PopupMenu(QtWidgets.QMenu, BaseSubcomponent):

    r"""A simple pop-up menu that is shown at the cursor position to obtain a
    user response.
    """
    def __init__(self, main_window, actions, cancel=True, title=None):
        r"""Constructor.

        Parameters
        ----------
        main_window : QWidget
            The main-window object.
        actions : list
            A list of actions, where each action is (response code, text, icon
            name) tuple.
        cancel : bool, optional
            Indicates whether a cancel option should be shown.
        """
        QtWidgets.QMenu.__init__(self, main_window)
        self.setup(main_window)

        if title is not None:
            action = QtWidgets.QAction(self.theme.qicon(u'dialog-information'),
                                       title, self)
            action._id = None
            action.setDisabled(True)
            self.addAction(action)
            self.addSeparator()

        for _id, text, icon in actions:
            action = QtWidgets.QAction(self.theme.qicon(icon), text, self)
            action._id = _id
            self.addAction(action)
        if cancel:
            self.addSeparator()
            action = QtWidgets.QAction(self.theme.qicon(u'edit-clear'),
                                       _(u'Cancel'), self)
            action._id = None
            self.addAction(action)

    def show(self):
        r"""Shows the pop-up menu.

        Returns
        -------
        A response code or None if the popup was cancelled.
        """
        action = self.exec_(QtGui.QCursor.pos())
        if action is None:
            return None
        return action._id


# Alias for backwards compatibility
popup_menu = PopupMenu



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp.backend import Backend, configurable
from libopensesame.exceptions import InvalidValue
import warnings


class Mouse(Backend):
    """The `Mouse` class is used to collect mouse input. You generally create a
    `Mouse` object with the `Mouse()` factory function, as described in the
    section [Creating a Mouse](#creating-a-mouse).
    
    __Example__
    
    ~~~ .python
    # Draw a 'fixation-dot mouse cursor' until a button is clicked
    my_mouse = Mouse()
    my_canvas = Canvas()
    while True:
        button, position, timestamp = my_mouse.get_click(timeout=20)
        if button is not None:
            break
        (x,y), time = my_mouse.get_pos()
        my_canvas.clear()
        my_canvas.fixdot(x, y)
        my_canvas.show()
    ~~~
    
    [TOC]
    
    ## Things to know
    
    ### Creating a Mouse
    
    You generally create a `Mouse` with the `Mouse()` factory function:
    
    ~~~ .python
    my_mouse = Mouse()
    ~~~
    
    Optionally, you can pass [Response keywords](#response-keywords) to `Mouse()`
    to set the default behavior:
    
    ~~~ .python
    my_mouse = Mouse(timeout=2000)
    ~~~
    
    ### Coordinates
    
    - When *Uniform coordinates* is set to 'yes', coordinates are relative to the
      center of the display. That is, (0,0) is the center. This is the default as
      of OpenSesame 3.0.0.
    - When *Uniform coordinates* is set to 'no', coordinates are relative to the
      top-left of the display. That is, (0,0) is the top-left. This was the default
      in OpenSesame 2.9.X and earlier.
    
    ### Button numbers
    
    Mouse buttons are numbered as follows:
    
    1. Left button
    2. Middle button
    3. Right button
    4. Scroll up
    5. Scroll down
    
    ### Touch screens
    
    When working with a touch screen, a touch is registered as button 1
    (left button).
    
    ### Response keywords
    
    Functions that accept `**resp_args` take the following keyword arguments:
    
    - `timeout` specifies a timeout value in milliseconds, or is set to `None` to
      disable the timeout.
    - `buttonlist` specifies a list of buttons that are accepted, or is set to
      `None` accept all buttons.
    - `visible` indicates whether the mouse cursor becomes visible when a click is
      collected (`True` or `False`). To immediately change cursor visibility, use
      `Mouse.show_cursor()`.
    
    ~~~ .python
    # Get a left or right button press with a timeout of 3000 ms
    my_mouse = Mouse()
    button, time = my_mouse.get_click(buttonlist=[1,3], timeout=3000)
    ~~~
    
    Response keywords only affect the current operation (except when passed to
    `Mouse()` when creating the object). To change the behavior for all subsequent
    operations, set the response properties directly:
    
    ~~~ .python
    # Get two left or right presses with a 5000 ms timeout
    my_mouse = Mouse()
    my_mouse.buttonlist = [1,3]
    my_mouse.timeout = 5000
    button1, time1 = my_mouse.get_click()
    button2, time2 = my_mouse.get_click()
    ~~~
    
    Or pass the response keywords to `Mouse()` when creating the object:
    
    ~~~ .python
    # Get two left or right presses with a 5000 ms timeout
    my_mouse = Mouse(buttonlist=[1,3], timeout=5000)
    button1, time1 = my_mouse.get_click()
    button2, time2 = my_mouse.get_click()
    ~~~
    """
    def __init__(self, experiment, **resp_args):
        r"""Constructor to create a new `Mouse` object. You do not generally
        call this constructor directly, but use the `Mouse()` function,
        which
        is described here: [/python/common/]().

        Parameters
        ----------
        experiment : experiment
            The experiment object.
        **resp_args : dict
            Optional [response keywords](#response-keywords) that will be used
            as the default for this `Mouse` object.


        Examples
        --------
        >>> my_mouse = Mouse(buttonlist=[1, 2], timeout=2000)
        """
        self.experiment = experiment
        self._cursor_shown = False
        Backend.__init__(self, configurables={
            u'timeout': self.assert_numeric_or_None,
            u'buttonlist': self.assert_list_or_None,
            u'visible': self.assert_bool,
        }, **resp_args
        )

    def set_config(self, **cfg):

        # Add synonyms to buttonlist
        if u'buttonlist' in cfg and isinstance(cfg[u'buttonlist'], list):
            try:
                cfg[u'buttonlist'] = \
                    [int(button) for button in cfg[u'buttonlist']]
            except:
                raise InvalidValue(f'buttonlist must be a list of numbers or '
                                   f'None, not {cfg["buttonlist"]}')
        Backend.set_config(self, **cfg)

    def default_config(self):

        return {
            u'timeout': None,
            u'buttonlist': None,
            u'visible': False
        }

    def show_cursor(self, show=True):
        r"""Immediately changes the visibility of the mouse cursor.

        __Note:__
        In most cases, you will want to use the `visible`
        keyword, which
        changes the visibility during response collection,
        that is, while
        `mouse.get_click()` is called. Calling 
        `show_cursor()` will not
        implicitly change the value of `visible`, 
        which can lead to the
        somewhat unintuitive behavior that the cursor
        is hidden as soon as
        `get_click()` is called.

        Parameters
        ----------
        show : bool, optional
            Indicates whether the cursor is shown (True) or hidden (False).
        """
        self._cursor_shown = show

    def set_pos(self, pos=(0, 0)):
        r"""Sets the position of the mouse cursor.

        __Warning:__ `set_pos()` is
        unreliable and will silently fail on
        some systems.

        Parameters
        ----------
        pos : tuple, optional
            An (x,y) tuple for the new mouse coordinates.

        Examples
        --------
        >>> my_mouse = Mouse()
        >>> my_mouse.set_pos(pos=(0,0))
        """
        pass

    @configurable
    def get_click(self, **resp_args):
        r"""Collects a mouse click.

        Parameters
        ----------
        **resp_args : dict
            Optional [response keywords](#response-keywords) that will be used
            for this call to `Mouse.get_click()`. This does not affect
            subsequent operations.


        Returns
        -------
        tuple
            A (button, position, timestamp) tuple. The button and position are
            `None` if a timeout occurs. Position is an (x, y) tuple in screen
            coordinates.

        Examples
        --------
        >>> my_mouse = Mouse()
        >>> button, (x, y), timestamp = my_mouse.get_click(timeout=5000)
        >>> if button is None:
        >>>         print('A timeout occurred!')
        """
        raise NotImplementedError()

    @configurable
    def get_click_release(self, **resp_args):
        r"""*New in v3.2.0*

        Collects a mouse-click release.

        *Important:* This
        function is currently not implemented for the
        *psycho* backend.

        Parameters
        ----------
        **resp_args : dict
            Optional [response keywords](#response-keywords) that will be used
            for this call to `Mouse.get_click_release()`. This does not affect
            subsequent operations.


        Returns
        -------
        tuple
            A (button, position, timestamp) tuple. The button and position are
            `None` if a timeout occurs. Position is an (x, y) tuple in screen
            coordinates.

        Examples
        --------
        >>> my_mouse = Mouse()
        >>> button, (x, y), timestamp = my_mouse.get_click_release(timeout=5000)
        >>> if button is None:
        >>>         print('A timeout occurred!')
        """
        raise NotImplementedError()

    def get_pos(self):
        r"""Returns the current position of the cursor.

        Returns
        -------
        tuple
            A (position, timestamp) tuple.

        Examples
        --------
        >>> my_mouse = Mouse()
        >>> (x, y), timestamp = my_mouse.get_pos()
        >>> print('The cursor was at (%d, %d)' % (x, y))
        """
        raise NotImplementedError()

    def get_pressed(self):
        r"""Returns the current state of the mouse buttons. A True value means
        the button is currently being pressed.

        Returns
        -------
        tuple.
            A (button1, button2, button3) tuple of boolean values.

        Examples
        --------
        >>> my_mouse = Mouse()
        >>> buttons = my_mouse.get_pressed()
        >>> b1, b2, b3 = buttons
        >>> print('Currently pressed mouse buttons: (%d,%d,%d)' % (b1,b2,b3))
        """
        raise NotImplementedError()

    def flush(self):
        r"""Clears all pending input, not limited to the mouse.

        Returns
        -------
        bool
            True if a button had been clicked (i.e., if there was something to
            flush) and False otherwise.

        Examples
        --------
        >>> my_mouse = Mouse()
        >>> my_mouse.flush()
        >>> button, position, timestamp = my_mouse.get_click()
        """
        raise NotImplementedError()

    def synonyms(self, button):
        r"""Gives a list of synonyms for a mouse button. For example, 1 and
        'left_button' are synonyms.

        Parameters
        ----------
        button : int, str, unicode
            A button value.

        Returns
        -------
        list
            A list of synonyms.
        """
        button_map = [
            (1, u"left_button"),
            (2, u"middle_button"),
            (3, u"right_button"),
            (4, u"scroll_up"),
            (5, u"scroll_down")
        ]
        for bm in button_map:
            if button in bm:
                return bm
        return []


# Non PEP-8 alias for backwards compatibility
mouse = Mouse



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.exceptions import BackendNotSupported
from libopensesame.misc import camel_case
from importlib import import_module
from . import _libjoystick


def LibJoystick(experiment, **kwargs):
    r"""A factory that returns a back-end specific joystick module.

    Parameters
    ----------
    experiment
        The experiment object.
    **kwargs : dict
        Keywords to be passed on the joystick init.
    """
    if experiment.var.canvas_backend == u'psycho':
        raise BackendNotSupported(
            'The joystick plug-in does not yet support the psycho back-end')
    backend = u'legacy'
    mod = import_module(f'{_libjoystick.__package__}.{backend}')
    cls = getattr(mod, camel_case(backend))
    return cls(experiment, **kwargs)



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame import metadata
from libqtopensesame.extensions import BaseExtension
from libqtopensesame.misc.config import cfg
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'opensesame_4_notifications', category=u'extension')


class opensesame_4_notifications(BaseExtension):

    r"""Provides tips for new users of OpenSesame 4."""
    def event_os4n_dismiss_old_experiment(self):
        r"""Permanently disables the old-experiment tab."""
        cfg.os4n_old_experiment_notification = False
        self.main_window.tabwidget.close_current()

    def event_open_experiment(self, path):
        r"""Opens a notification tab when opening an experiment that was
        created with an older version of OpenSesame.

        Parameters
        ----------
        path : unicode
            The full path of the experiment file.
        """
        exp_api = self.experiment.front_matter[u'API']
        # If the major API of the experiment is the same, and the minor API is
        # the same or older
        if exp_api.version[0] == metadata.api.version[0] and \
                exp_api.version[1] <= metadata.api.version[1]:
            return
        self.main_window.current_path = None
        self.main_window.window_message(u'New experiment')
        self.main_window.set_unsaved(True)
        if not cfg.os4n_old_experiment_notification:
            return
        src = u'old-experiment.md' if exp_api < metadata.api \
            else u'new-experiment.md'
        self.tabwidget.open_markdown(self.ext_resource(src))



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.exceptions import InvalidValue, AbortCoroutines


class BaseTask:

    r"""A task controls the coroutine for one item."""
    ALIVE = 0
    DEAD = 1
    ABORT = 2

    def __init__(self, coroutines, start_time, end_time, abort_on_end=False):
        r"""Constructor.

        Parameters
        ----------
        start_time : int
            The start-time of the coroutine.
        end_time : int
            The end-time of the coroutine.
        abort_on_end : bool`, optional
            Indicates whether an ABORT signal should be given when the task
            ends.
        """
        if not (isinstance(start_time, (int, float)) and start_time >= 0) or \
                not (isinstance(end_time, (int, float)) and end_time >= start_time):
            raise InvalidValue(f'Start (now: {start_time}) and end (now: '
                               f'{end_time}) times need to be non-negative '
                               f'numeric values and end time needs to be '
                               f'equal to higher than start time')
        self.start_time = start_time
        self.end_time = end_time
        self.coroutines = coroutines
        self._abort_on_end = abort_on_end

    def started(self, dt):
        """
        arguments:
                dt:
                        desc:	The current time relative to onset of the coroutines.
                        type:	int

        returns:
                desc:	True if the current item is started, False otherwise.
                type:	bool
        """
        return dt >= self.start_time

    def stopped(self, dt):
        r"""Checks whether an item is stopped, and sends the stop signal to the
        item if it should be stopped.

        Parameters
        ----------
        dt : int
            The current time relative to onset of the coroutines.

        Returns
        -------
        bool
            True if the current item is started, False otherwise.
        """
        if dt < self.end_time:
            return False
        try:
            self.coroutine.send(False)
        except StopIteration:
            self.coroutines.event('stopped %s' % self.coroutine)
        return True

    def kill(self):
        r"""Sends the stop signal to an item."""
        try:
            self.coroutine.send(False)
        except StopIteration:
            self.coroutines.event('killed %s' % self.coroutine)

    def launch(self):
        r"""Launches an item. All items are launched at coroutine start."""
        raise NotImplementedError()

    def step(self):
        r"""Lets the item yield one cycle."""
        try:
            if self.coroutine.send(True) is False:
                return self.ABORT
        except AbortCoroutines:
            return self.ABORT
        except StopIteration:
            self.coroutines.event('died %s' % self.coroutine)
            return self.ABORT if self._abort_on_end else self.DEAD
        return self.ALIVE



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.

# About

This module is used to maintain configuration settings. When this module is
first loaded, a single instance of the `config` class is instantiated, whicn is
subsequently used for all configuration getting and setting (i.e. a singleton
design pattern).

	from libqtopensesame.misc.config import cfg
	cfg.my_setting = 'my_value' # set
	print(cfg.my_setting) # get
"""
from libopensesame.py3compat import *
from qtpy import QtCore
import libopensesame.misc
from libopensesame.oslogging import oslogger
import platform
import sys


DEFAULT_CONFIG = {
    "cfg_ver": 0,
    "_initial_window_geometry": QtCore.QByteArray(),
    "_initial_window_state": QtCore.QByteArray(),
    "auto_update_check": True,
    "default_logfile_folder": libopensesame.misc.home_folder(),
    "default_pool_folder": libopensesame.misc.home_folder(),
    "disabled_plugins": "",
    "disabled_extensions": "",
    "file_dialog_path": "",
    "file_pool_size_warning": 104857600,
    "loop_wizard": None,
    "onetabmode": True,
    "quick_run_logfile": u"quickrun.csv",
    "recent_files": u"",
    "reset_console_on_experiment_start": True,
    "shortcut_itemtree": u"Ctrl+1",
    "shortcut_tabwidget": u"Ctrl+2",
    "shortcut_stdout": u"Ctrl+3",
    "shortcut_pool": u"Ctrl+4",
    "shortcut_copy_clipboard_unlinked": u"Ctrl+C",
    "shortcut_copy_clipboard_linked": u"Ctrl+Shift+C",
    "shortcut_paste_clipboard": u"Ctrl+V",
    "shortcut_delete": u"Del",
    "shortcut_permanently_delete": u"Shift+Del",
    "shortcut_context_menu": u"+",
    "shortcut_rename": u"F2",
    "shortcut_edit_runif": u"F3",
    "style": u"",
    "theme": u"default",
    "toolbar_size": 32,
    "toolbar_text": False,
    "runner": u"multiprocess",
    "opensesamerun_exec": u"",
    "start_drag_delay": 300,
    "pos": QtCore.QPoint(200, 200),
    "size": QtCore.QSize(1000, 600),
    "url_website": u"http://osdoc.cogsci.nl/",
    "url_facebook": u"http://www.facebook.com/cognitivescience",
    "url_twitter": u"http://www.twitter.com/cogscinl",
    'sketchpad_placeholder_color': u'#00FF00',
    'sketchpad_grid_color': u'#00FF00',
    'sketchpad_grid_thickness_thin': 2,
    'sketchpad_grid_thickness_thick': 4,
    'sketchpad_grid_opacity': 32,
    'sketchpad_preview_color': u'#00FF00',
    'sketchpad_preview_penwidth': 2,
    'mode': u'default',
    'locale': u'',
}
DEFAULT_CONFIG_LINUX = {}
DEFAULT_CONFIG_MAC = {}
DEFAULT_CONFIG_WINDOWS = {}


class Config:

    def __init__(self):
        self.reset()

    def __str__(self):
        s = u''
        for key, val in self.config.items():
            if not isinstance(val, str) and not isinstance(val, int) \
                    and not isinstance(val, float):
                val = type(val)
            s += u'%s: %s\n' % (key, val)
        return s

    def __getattr__(self, setting):
        r"""A getter for settings, to allow for easy access

        Parameters
        ----------
        setting
            The setting to get
        """
        if setting not in self.config:
            raise KeyError(f'The setting {setting} does not exist')
        return self.config[setting]

    def __setattr__(self, setting, value):
        """A setter for settings, to allow for easy access"""
        if setting not in self.config:
            raise KeyError(f'The setting {setting} does not exist')
        self.config[setting] = value
        self.config[u'cfg_ver'] += 1

    def __setitem__(self, setting, value):
        r"""Emulate dict API."""
        self.__setattr__(setting, value)

    def __getitem__(self, setting):
        r"""Emulate dict API."""
        return self.__getattr__(setting)

    def __contains__(self, setting):
        """
        returns:
                True if setting exists, False otherwise.
        """
        return setting in self.config

    def register(self, setting, default):
        r"""Registers a new setting, if it doesn't already exist.

        Parameters
        ----------
        setting
            The setting name.
        default
            A default value, which is used if the setting doesn't already
            exist.
        """
        qsettings = QtCore.QSettings(u"cogscinl", self.mode)
        qsettings.beginGroup(u"MainWindow")
        if '--start-clean' in sys.argv:
            self.config[setting] = default
        elif setting not in self.config:
            value = qsettings.value(setting, default)
            value = self.type_qvariant(value, default)
            self.config[setting] = value
        qsettings.endGroup()

    def parse_cmdline_args(self, args):
        """Apply settings that were specified on the command line. The expected
        format is as follows: [name]=[val];[name]=[val];...

        Parameters
        ----------
        args: str
            The string of command line arguments
        """
        if args is None:
            return
        for arg in args.split(u";"):
            a = arg.split(u"=")
            if len(a) == 2:
                # Automagically determine the data type
                if a[1] == u"True":
                    val = True
                elif a[1] == u"False":
                    val = False
                else:
                    try:
                        val = int(a[1])
                    except:
                        try:
                            val = float(a[1])
                        except:
                            val = a[1]
                # Apply the argument
                try:
                    self.__setattr__(a[0], val)
                    oslogger.debug(u"%s = %s" % (a[0], val))
                except:
                    oslogger.error(u"Failed to parse argument: %s" % arg)

    def type_qvariant(self, value, default=None):
        r"""Typecasts a value to a normal type that matches the type of a
        default value, because QSettings returns str objects.

        Parameters
        ----------
        value
            A value.
        default
            A default value.

        Returns
        -------
        A value.
        """
        if not isinstance(value, str):
            return value
        if (default is None or isinstance(default, bool)):
            if value == 'true':
                return True
            if value == 'false':
                return False
        if (default is None or isinstance(default, int)):
            try:
                return int(value)
            except (TypeError, ValueError):
                pass
        if (default is None or isinstance(default, float)):
            try:
                return float(value)
            except (TypeError, ValueError):
                pass
        return value

    def restore(self, mode):
        r"""Restore settings from a QSettings."""
        oslogger.debug('restoring config profile {}'.format(mode))
        qsettings = QtCore.QSettings(u"cogscinl", mode)
        qsettings.beginGroup('OpenSesame4')
        for setting, default in self.config.items():
            try:
                value = qsettings.value(setting, default)
            except TypeError:
                continue
            value = self.type_qvariant(value, default)
            self.config[setting] = value
        for setting in qsettings.allKeys():
            if setting in self.config:
                continue
            oslogger.debug(f'unregistered setting {setting}')
            value = qsettings.value(setting)
            value = self.type_qvariant(value, None)
            self.config[setting] = value
        self.mode = mode
        qsettings.endGroup()

    def save(self):
        r"""Save settings to a QSettings."""
        oslogger.debug('saving config profile {}'.format(self.mode))
        qsettings = QtCore.QSettings(u"cogscinl", self.mode)
        qsettings.beginGroup('OpenSesame4')
        for setting, value in self.config.items():
            if setting != u"cfg_ver":
                qsettings.setValue(setting, value)
        qsettings.endGroup()

    def clear(self):
        r"""Clears all settings."""
        qsettings = QtCore.QSettings(u"cogscinl", self.mode)
        qsettings.beginGroup('OpenSesame4')
        qsettings.clear()
        qsettings.endGroup()

    def version(self):
        r"""Gets the current version of the config."""
        return self.cfg_ver

    def reset(self):
        r"""Resets the configution back to default."""
        object.__setattr__(self, u'config', {})
        self.config.update(DEFAULT_CONFIG)
        if platform.system() == u"Windows":
            self.config.update(DEFAULT_CONFIG_WINDOWS)
        elif platform.system() == u"Darwin":
            self.config.update(DEFAULT_CONFIG_MAC)
        elif platform.system() == u"Linux":
            self.config.update(DEFAULT_CONFIG_LINUX)

    def nuke(self):
        r"""Clears the config."""
        self.reset()
        self.clear()
        self.save()


# Create a singleton config instance
cfg = Config()



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from pygame.locals import *
import pygame
from openexp._sampler.sampler import Sampler
from libopensesame.exceptions import SoundFileDoesNotExist, \
    UnsupportedSoundFileFormat, InvalidValue
from libopensesame.oslogging import oslogger
from libopensesame import misc
from openexp.keyboard import Keyboard
from openexp.backend import configurable
import os.path
try:
    import numpy
except:
    numpy = None
import pygame.mixer as mixer


class Legacy(Sampler):

    r"""This is a sampler backend built on top of PyGame. For function
    specifications and docstrings, see `openexp._sampler.sampler`.
    """
    # The settings variable is used by the GUI to provide a list of back-end
    # settings
    settings = {
        u"sound_buf_size": {
            u"name": u"Sound buffer size",
            u"description": u"Size of the sound buffer (increase if playback is choppy)",
            u"default": 1024
        },
        u"sound_freq": {
            u"name": u"Sampling frequency",
            u"description": u"Determines the sampling rate",
            u"default": 48000
        },
        u"sound_sample_size": {
            u"name": u"Sample size",
            u"description": u"Determines the bith depth (negative = signed)",
            u"default": -16
        },
        "sound_channels": {
            u"name": u"The number of sound channels",
            u"description": u"1 = mono, 2 = stereo",
            u"default": 2
        },
    }

    def __init__(self, experiment, src, **playback_args):

        if src is not None:
            if isinstance(src, str):
                if not os.path.exists(src):
                    raise SoundFileDoesNotExist(src)
                if os.path.splitext(src)[1].lower() not in (".ogg", ".wav"):
                    raise UnsupportedSoundFileFormat(src)
            self.sound = mixer.Sound(src)
        Sampler.__init__(self, experiment, src, **playback_args)
        self.keyboard = Keyboard(experiment)
        self._channel = None

    def set_config(self, **cfg):

        if u'duration' in cfg and cfg[u'duration'] is None:
            cfg[u'duration'] = 0
        if u'fade_in' in cfg and cfg[u'fade_in'] is None:
            cfg[u'fade_in'] = 0
        Sampler.set_config(self, **cfg)
        if u'volume' in cfg:
            self.sound.set_volume(cfg[u'volume'])
        if u'pitch' in cfg:
            self.adjust_pitch(cfg[u'pitch'])
        if u'pan' in cfg:
            self.adjust_pan(cfg[u'pan'])

    def adjust_pitch(self, p):

        # On Android, numpy does not exist and this is not supported
        if numpy is None:
            return
        if type(p) not in (int, float) or p <= 0:
            raise InvalidValue(f'pitch should be a positive number, not {p}')
        if p == 1:
            return
        buf = pygame.sndarray.array(self.sound)
        _buf = []
        for i in range(int(float(len(buf)) / p)):
            _buf.append(buf[int(float(i) * p)])
        self.sound = pygame.sndarray.make_sound(
            numpy.array(_buf, dtype=u"int16"))

    def adjust_pan(self, p):

        # On Android, numpy does not exist and this is not supported
        if numpy is None:
            return
        if type(p) not in (int, float) and p not in (u"left", u"right"):
            raise InvalidValue(
                f'pan should be a number, "left" or "right", not {p}')
        if p == 0:
            return
        buf = pygame.sndarray.array(self.sound)
        for i in range(len(buf)):
            l = buf[i][0]
            r = buf[i][1]
            if p == "left":
                r = 0
            elif p == "right":
                l = 0
            elif p < 0:
                r = int(float(r) / abs(p))
            else:
                l = int(float(l) / p)
            buf[i][0] = l
            buf[i][1] = r
        self.sound = pygame.sndarray.make_sound(numpy.array(buf))

    @configurable
    def play(self, **playback_args):

        self._channel = self.sound.play(maxtime=self.duration,
                                        fade_ms=self.fade_in)
        if self.block:
            self.wait()

    def stop(self):

        if self._channel is None:
            return
        self._channel.stop()
        
    def pause(self):

        if self._channel is None:
            return
        self._channel.pause()

    def resume(self):

        if self._channel is None:
            return
        self._channel.unpause()

    def is_playing(self):

        if self._channel is None:
            return
        return self._channel.get_busy()

    def wait(self):

        if self._channel is None:
            return
        while self._channel.get_busy():
            self.keyboard.flush()

    @staticmethod
    def init_sound(experiment):

        oslogger.info(
            u"sampling freq = %d, buffer size = %d"
            % (experiment.var.sound_freq, experiment.var.sound_buf_size)
        )
        if hasattr(mixer, u'get_init') and mixer.get_init():
            oslogger.warning(u'mixer already initialized, closing')
            pygame.mixer.quit()
        mixer.pre_init(experiment.var.sound_freq,
                       experiment.var.sound_sample_size,
                       experiment.var.sound_channels,
                       experiment.var.sound_buf_size)
        try:
            mixer.init()
        except pygame.error:
            oslogger.error(u'failed to initialize mixer')

    @staticmethod
    def close_sound(experiment):

        mixer.quit()


# Non PEP-8 alias for backwards compatibility
legacy = Legacy



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.widgets._label import Label


class Button(Label):

    r"""The `Button` widget is a clickable text string, by default surrounded
    by
    a button-like frame.

    __Example (OpenSesame script):__

    ~~~
    widget 0 0 1
    1 button text='Click me!' center='yes' frame='yes' var='response'
    ~~~
    Defining a button widget with Python inline code:

    __Example (Python):__
    ~~~ .python
    form = Form()
    button = Button(text='Click me!', frame=True,
    center=True, var='response')
    form.set_widget(button, (0,0))
    form._exec()
    ~~~

    [TOC]
    """
    def __init__(self, form, text=u'button', frame=True, center=True, var=None):
        r"""Constructor to create a new `Button` object. You do not generally
        call this constructor directly, but use the `Button()` factory
        function, which is described here: [/python/common/]().

        Parameters
        ----------
        form : form
            The parent form.
        text : str, unicode, optional
            Button text.
        frame : bool, optional
            Indicates whether a frame should be drawn around the widget.
        center : bool, optional
            Indicates whether the text should be centered.
        var : str, unicode, NoneType, optional
            The name of the experimental variable that should be used to log
            the widget status.
        """
        Label.__init__(self, form, text, frame=frame, center=center)
        self.type = u'button'
        self.var = var
        self.set_var(False)

    def on_mouse_click(self, pos):
        r"""Is called when the user clicks on the button. Returns the button
        text.

        Parameters
        ----------
        pos : tuple
            An (x, y) coordinates tuple.

        Returns
        -------
        unicode
            The button text.
        """
        self.theme_engine.click()
        self.set_var(True)
        return self.text


button = Button



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.widgets.tree_base_item import TreeBaseItem
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'tree_recursion_error_item', category=u'core')


class TreeRecursionErrorItem(TreeBaseItem):

    r"""A placeholder item that is shown when there is a recursion in the
    experiment that prevents the item tree from being built.
    """
    def __init__(self, main_window):
        r"""Constructor.

        Parameters
        ----------
        main_window : qtopensesame
            The main-window object.
        """
        super().__init__()
        self.setup(main_window)
        self.setText(0, _(u'Recursion detected'))
        self.setIcon(0, self.theme.qicon(u'dialog-error'))
        self._droppable = False
        self._draggable = False
        self.name = u'__recursion_error__'

    def open_tab(self):

        self.main_window.tabwidget.open_general_script()


# Alias for backwards compatibility
tree_recursion_error_item = TreeRecursionErrorItem



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp._canvas._arrow.arrow import Arrow
from openexp._canvas._polygon.legacy import Legacy as LegacyPolygon


class Legacy(LegacyPolygon, Arrow):

    def copy(self, canvas):

        vertices = self._properties['vertices']
        del self._properties['vertices']
        arrow_copy = LegacyPolygon.copy(self, canvas)
        self._properties['vertices'] = vertices
        return arrow_copy

    def _on_attribute_change(self, **kwargs):

        # When a change occurs the shapely line (if any) needs to be
        # redetermined in __contains__
        self._properties['vertices'] = self._shape(
            self.sx,
            self.sy,
            self.ex,
            self.ey,
            self.body_length,
            self.body_width,
            self.head_width
        )
        Arrow._on_attribute_change(self, **kwargs)



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import os
import tarfile
import tempfile
import shutil
from libopensesame.osexpfile import OSExpBase
from libopensesame.oslogging import oslogger


class OSExpWriter(OSExpBase):

    r"""A writer of osexp files. The format (plain text or tar.gz) depends on
    whether the file pool contains files.
    """
    def __init__(self, exp, path):
        r"""Constructor.

        Parameters
        ----------
        exp : experiment
            An experiment object.
        path : str
            A path to the experiment file.
        """
        super().__init__(exp)
        self._path = path
        self._experiment_path = os.path.dirname(path)
        if self.format == 'scriptfile':
            self._write_scriptfile()
        elif self.format == 'tarfile':
            self._write_tarfile()
        else:
            raise TypeError(u'Invalid format')

    @property
    def script(self):
        """See osexpbase"""
        return self._exp.to_string()

    def _determine_format(self):
        """See osexpbase"""
        return 'scriptfile' if not self._pool.count_included() else 'tarfile'

    def _write_scriptfile(self):
        r"""Writes a plain-text file."""
        with safe_open(self._path, u'w') as fd:
            fd.write(self.script)

    def _write_tarfile(self):
        r"""Writes a .tar.gz file."""
        # Write the script to a text file in the file pool
        script_path = os.path.join(self._pool.folder(), u'script.opensesame')
        with safe_open(script_path, u'w') as fd:
            fd.write(self.script)
        # Create the archive in a a temporary folder and move it afterwards.
        # This hack is needed, because tarfile fails on a Unicode path.
        tmp_path = tempfile.mktemp(suffix=u'.osexp')
        tar = tarfile.open(tmp_path, u'w:gz')
        tar.add(script_path, u'script.opensesame')
        os.remove(script_path)
        # We also create a temporary pool folder, where all the filenames are
        # Unicode sanitized to ASCII format. Again, this is necessary to deal
        # with poor Unicode support in .tar.gz.
        tmp_pool = tempfile.mkdtemp(suffix=u'.opensesame.pool')
        for fname in os.listdir(self._pool.folder()):
            src = os.path.join(self._pool.folder(), fname)
            if not os.path.isfile(src):
                oslogger.warning('{} is not a file'.format(src))
                continue
            dst = os.path.join(tmp_pool, self._syntax.to_ascii(fname))
            shutil.copyfile(src, dst)
        tar.add(tmp_pool, u'pool', True)
        tar.close()
        # Move the file to the intended location
        shutil.move(tmp_path, self._path)
        # Clean up the temporary pool folder
        try:
            shutil.rmtree(tmp_pool)
        except:
            pass



# -*- coding:utf-8 -*-

"""
This file is part of openexp.

openexp is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

openexp is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with openexp.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.misc.config import cfg
from libqtopensesame.sketchpad_elements._base_element import BaseElement


class BaseRectEllipse(BaseElement):

    pos_start = None
    preview = None

    @classmethod
    def mouse_release(cls, sketchpad, pos):

        if cls.pos_start is None:
            return
        w = pos[0]-cls.pos_start[0]
        h = pos[1]-cls.pos_start[1]
        # Don't draw line-like shapes
        if abs(w) < 1 or abs(h) < 1:
            cls.pos_start = None
            sketchpad.canvas.removeItem(cls.preview)
            return
        properties = {
            u'x':		cls.pos_start[0],
            u'y':		cls.pos_start[1],
            u'w':		w,
            u'h':		h,
            u'color': 	sketchpad.current_color(),
            u'penwidth': sketchpad.current_penwidth(),
            u'fill': sketchpad.current_fill(),
            u'show_if': sketchpad.current_show_if()
        }
        e = cls(sketchpad, properties=properties)
        cls.pos_start = None
        sketchpad.canvas.removeItem(cls.preview)
        return e

    @classmethod
    def mouse_move(cls, sketchpad, pos):

        if cls.pos_start is None or cls.preview is None:
            return
        x = cls.pos_start[0]
        y = cls.pos_start[1]
        w = pos[0]-cls.pos_start[0]
        h = pos[1]-cls.pos_start[1]
        if w < 0:
            x += w
            w *= -1
        if h < 0:
            y += h
            h *= -1
        cls.preview.setRect(x, y, w, h)

    @classmethod
    def reset(cls):

        cls.pos_start = None

    @staticmethod
    def requires_color():
        return True

    @staticmethod
    def requires_penwidth():
        return True

    @staticmethod
    def requires_fill():
        return True


# Alias for backwards compatibility
base_rect_ellipse = BaseRectEllipse



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *

from libopensesame.widgets._button import Button
from libopensesame.widgets._checkbox import Checkbox
from libopensesame.widgets._form import Form
from libopensesame.widgets._image import ImageWidget
from libopensesame.widgets._image_button import ImageButton
from libopensesame.widgets._label import Label
from libopensesame.widgets._rating_scale import RatingScale
from libopensesame.widgets._text_input import TextInput
from libopensesame.widgets._widget import Widget

# alias for backwards compatibility
button = Button
checkbox = Checkbox
form = Form
image = ImageWidget
image_button = ImageButton
label = Label
rating_scale = RatingScale
text_input = TextInput
widget = Widget



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.extensions import BaseExtension
from libqtopensesame.misc import display
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'system_check', category=u'extension')


class SystemCheck(BaseExtension):

    def event_startup(self):
        r"""Performs several checks when OpenSesame is started. For now, it
        only checks whether the display is scaled or not, and if so, provides a
        warning because this may affect how stimuli look during the experiment.
        """
        if display.display_scaling != 1:
            self.extension_manager.fire(
                u'notify',
                message=_(
                    u'Display scaling is set to %d%% by the operating system. '
                    u'This may affect the appearance of your experiment.'
                ) % (100*display.display_scaling),
                category=u'warning',
                timeout=5000
            )



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from .variable_inspector_cell import VariableInspectorCell
from libqtopensesame.widgets.base_widget import BaseWidget
from libqtopensesame.misc import drag_and_drop
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'variable_inspector', category=u'extension')


class VariableInspectorWidget(BaseWidget):
    r"""The variable inspector widget, which includes the table, filter button,
    etc.
    """
    
    def __init__(self, main_window, ext):
        r"""Constructor.

        Parameters
        ----------
        main_window
            The main-window object.
        ext
            The variable_inspector-extension object.
        """
        super().__init__(
            main_window,
            ui=u'extensions.variable_inspector.variable_inspector')
        self.ext = ext
        self.ui.edit_variable_filter.textChanged.connect(self.refresh)
        self.ui.button_help_variables.clicked.connect(self.ext.open_help)
        self.ui.button_reset.clicked.connect(self._reset)
        self.ui.table_variables.mousePressEvent = self.start_drag
        self._workspace_globals = {}
        self.refresh()

    def focus(self):
        r"""Sets the focus to the filter box."""
        self.ui.edit_variable_filter.setFocus()

    def _reset(self):
        r"""Resets the console, to reset the workspace."""
        self.set_workspace_globals({})
        self.refresh()

    def var(self):
        """Returns a (var_store, alive) tuple, where alive indicates whether
        the var_store is from an active or finished experiment, or from an 
        inactive GUI experiment.
        """
        if (
                u'var' in self._workspace_globals and
                hasattr(self._workspace_globals[u'var'], u'inspect')
        ):
            return self._workspace_globals[u'var'], True
        return self.experiment.var, False

    def set_workspace_globals(self, global_dict):
        r"""Sets the workspace global dict.

        Parameters
        ----------
        global_dict
            The workspace global dict
        """
        self._workspace_globals = global_dict

    def start_drag(self, e):
        r"""Starts a variable drag operation.

        Parameters
        ----------
        e : QMousePressEvent
        """
        item = self.ui.table_variables.itemAt(e.pos())
        if item is None:
            return
        row = self.ui.table_variables.row(item)
        var = self.ui.table_variables.item(row, 0).text()
        drag_and_drop.send(self.ui.table_variables, var)

    def refresh(self):
        r"""Refreshes the table."""
        # Remember the view position
        scrollpos = self.ui.table_variables.verticalScrollBar().sliderPosition()
        col = self.ui.table_variables.currentColumn()
        row = self.ui.table_variables.currentRow()
        filt = str(self.ui.edit_variable_filter.text())
        var_store, alive = self.var()
        self.ui.label_no_heartbeat.setVisible(
            not self.main_window.runner_cls.has_heartbeat())
        self.ui.label_status.setText(
            _(u'Experiment status: <b>%s</b>') % self.main_window.run_status()
        )
        if alive and self.main_window.run_status() == u'finished':
            self.ui.widget_reset_message.show()
        else:
            self.ui.widget_reset_message.hide()
        # Filter the variables if necessary
        if len(filt) > 1:
            d = {
                var: info
                for var, info in var_store.inspect().items()
                if any(
                    q.strip() in var or
                    q.strip() in safe_decode(info[u'value'],
                                             errors=u'ignore') or
                    q.strip() in u' '.join(info[u'source'])
                    for q in filt.split(u'|')
                )
            }
        else:
            d = var_store.inspect()
        # Populate the table
        self.ui.table_variables.setRowCount(len(d))
        for i, var in enumerate(sorted(d.keys())):
            info = d[var]
            self.ui.table_variables.setItem(
                i, 0, VariableInspectorCell(var, info))
            self.ui.table_variables.setItem(
                i, 1, VariableInspectorCell(info['value'], info))
            self.ui.table_variables.setItem(
                i, 2, VariableInspectorCell(info['source'], info))
        # Restore the view position
        self.ui.table_variables.setCurrentCell(row, col)
        self.ui.table_variables.verticalScrollBar().setSliderPosition(
            scrollpos)



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame import metadata
from libopensesame.py3compat import *
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
import re
import os
from libopensesame.oslogging import oslogger
try:
    import markdown
    from markdown.extensions import attr_list, extra, toc
except ImportError:
    oslogger.error(u'Unable to import markdown, proceeding without markdown')
    markdown = None
try:
    from pygments import highlight
    from pygments.lexers import Python3TracebackLexer as TracebackLexer
    from pygments.lexers import Python3Lexer as PythonLexer
    from pygments.formatters import HtmlFormatter
except ImportError:
    highlight = None
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'markdown', category=u'core')


class MarkdownParser(BaseSubcomponent):

    r"""A Markdown parser with syntax highlighting."""
    def __init__(self, main_window):
        r"""Constructor.

        Parameters
        ----------
        main_window
            The main-window object.
        """
        self.setup(main_window)
        self.css = u'<style type="text/css">'
        with safe_open(self.main_window.theme.resource(u'markdown.css')) as fd:
            self.css += fd.read() % {u'background_image':
                                     os.path.abspath(self.main_window.theme.resource(
                                         u'background.png'))}
        if highlight is not None:
            self.traceback_lexer = TracebackLexer()
            self.python_lexer = PythonLexer()
            self.html_formatter = HtmlFormatter()
            self.css += self.html_formatter.get_style_defs(u'.highlight')
            self.re_script = re.compile(
                r'^~~~\s*.(?P<syntax>\w+)(?P<script>.*?)^~~~', re.S | re.M)
        self.css += u'</style><link href="https://fonts.googleapis.com/css?family=Roboto+Slab&display=swap" rel="stylesheet">'
        if markdown is not None:
            self.ext = [attr_list.AttrListExtension(), extra.ExtraExtension(),
                        toc.TocExtension(title=u'Overview'),
                        u'markdown.extensions.tables']
        self.footer = u'''
<p>
<a class="dismiss-button" href="opensesame://action.close_current_tab">%s</a>
</p>

<div class="footer">
%s
Copyright <a href="http://www.cogsci.nl/smathot">Sebastiaan Mathôt</a> 2010-2024
</div>
''' % (_(u'Dismiss this message'), metadata.identity)

    def highlight(self, md):
        r"""Replaces ~~~ blocks with syntax-highlighted HTML code.

        Parameters
        ----------
        md : str
            A Markdown  string.

        Returns
        -------
        str
            A Markdown  string.
        """
        if highlight is None:
            return md
        while True:
            m = re.search(self.re_script, md)
            if m is None:
                break
            orig = m.group()
            syntax = m.group(u'syntax')
            script = m.group(u'script')
            if syntax == u'traceback':
                lexer = self.traceback_lexer
            elif syntax == u'python':
                lexer = self.python_lexer
            else:
                md = md.replace(orig, u'<code>%s</code>\n' % script)
                continue
            new = highlight(script, lexer, self.html_formatter)
            md = md.replace(orig, new)
        return md

    def to_html(self, md):
        r"""Converts Markdown to HTML.

        Parameters
        ----------
        md : str
            A Markdown  string.

        Returns
        -------
        str
            A Markdown  string.
        """
        md = self.highlight(md)
        if markdown is None:
            return u'<pre>%s</pre>' % md
        html = markdown.markdown(md, extensions=self.ext, errors=u'ignore') \
            + self.css + self.footer
        if html.startswith(u'<p>title:'):
            title, body = tuple(html.split(u'\n', 1))
            html = u'<h1>%s</h1>\n\n%s' % (title[9:-4], body)
        return html


# Alias for backwards compatibility
markdown_parser = MarkdownParser



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
from qtpy.QtWidgets import QPushButton
from libqtopensesame.misc.translate import translation_context
from libopensesame.misc import snake_case
_ = translation_context(u'sketchpad', category=u'item')


class SketchpadElementButton(BaseSubcomponent, QPushButton):

    r"""A element-tool button for the sketchpad widget."""
    def __init__(self, sketchpad_widget, element):
        r"""Constructor.

        Parameters
        ----------
        sketchpad_widget : sketchpad-widget
        element : base_element
        """
        super().__init__(sketchpad_widget)
        self.element = element
        self.sketchpad_widget = sketchpad_widget
        self.setup(sketchpad_widget)
        self.setFlat(True)
        self.setCheckable(True)
        self.setIcon(self.theme.qicon(u'os-%s' % self.name))
        self.setToolTip(_(u'Draw %s element') % self.name)
        self.clicked.connect(self.select)

    def select(self, checked):
        r"""Select this element tool."""
        self.sketchpad_widget.unselect_all_tools()
        self.setChecked(True)
        self.sketchpad_widget.select_element_tool(self.element)

    @property
    def name(self):
        return snake_case(self.element.__name__)


# Alias for backwards compatibility
sketchpad_element_button = SketchpadElementButton



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtCore, QtWidgets
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'item_view_button', category=u'core')


class ItemViewButton(BaseSubcomponent, QtWidgets.QPushButton):

    r"""The view button in the top-right of the edit controls."""
    def __init__(self, item):
        r"""Constructor.

        Parameters
        ----------
        item : qtitem.
            An item.
        """
        super().__init__(item.main_window)
        self.item = item
        self.setup(item)
        self.set_view_icon(u'controls')
        self.setIconSize(QtCore.QSize(16, 16))
        self.menu_view = QtWidgets.QMenu()
        self.menu_view.addAction(self.view_controls_icon(),
                                 _(u'View controls'),
                                 self.item.set_view_controls)
        self.menu_view.addAction(self.view_script_icon(), _(u'View script'),
                                 self.item.set_view_script)
        self.menu_view.addAction(self.view_split_icon(), _(u'Split view'),
                                 self.item.set_view_split)
        self.setMenu(self.menu_view)
        self.setToolTip(_(u'Select view'))

    def view_controls_icon(self):
        """
        returns:
                desc:	The icon for the controls view.
                type: 	QIcon
        """
        return self.theme.qicon(u'os-view-controls')

    def view_script_icon(self):
        """
        returns:
                desc:	The icon for the script view.
                type: 	QIcon
        """
        return self.theme.qicon(u'os-view-script')

    def view_split_icon(self):
        """
        returns:
                desc:	The icon for the split view.
                type: 	QIcon
        """
        return self.theme.qicon(u'os-view-split')

    def set_view_icon(self, view):
        """
        desc:
                Sets the icon of the button according to a view.

        arguments:
                view:
                        desc:	'controls', 'script', or 'split'.
                        type:	unicode
        """
        if view == u'controls':
            self.setIcon(self.view_controls_icon())
        elif view == u'script':
            self.setIcon(self.view_script_icon())
        else:
            self.setIcon(self.view_split_icon())


# Alias for backwards compatibility
item_view_button = ItemViewButton



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtCore, QtWidgets
from collections import OrderedDict
from libopensesame.oslogging import oslogger
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
from libqtopensesame.widgets.toolbar_items_label import ToolbarItemsLabel
from libqtopensesame.widgets.toolbar_items_item import ToolbarItemsItem
from libqtopensesame.misc.translate import translation_context
from libqtopensesame.misc.config import cfg
_ = translation_context(u'toolbar_items', category=u'core')


class ToolbarItems(BaseSubcomponent, QtWidgets.QToolBar):
    """The item toolbar, which allows you to insert items into the experiment
    through drag and drop.
    
    Parameters
    ----------
    parent : QWidget
        The parent.
    """
    def __init__(self, parent):
        super().__init__(parent)
        self.setup(parent)
        self.orientationChanged.connect(self.build)
        self._items = []
        for child in self.children():
            if isinstance(child, QtWidgets.QToolButton):
                self._expand_button = child
                break

    def add_content(self, content):
        r"""Add double rows of content to the toolbar.

        Parameters
        ----------
        content : list
            A list of content widgets.
        """
        self._items += content
        for i, c in enumerate(content):
            if not i % 2:
                if i > 0:
                    self.addWidget(w)
                if self.orientation() == QtCore.Qt.Horizontal:
                    l = QtWidgets.QVBoxLayout()
                    l.setContentsMargins(6, 6, 6, 6)
                else:
                    l = QtWidgets.QHBoxLayout()
                    l.setContentsMargins(6, 6, 6, 6)
                l.setSpacing(12)
                w = QtWidgets.QWidget()
                w.setLayout(l)
            l.addWidget(c)
        if not i % 2:
            l.addStretch()
        self.addWidget(w)

    def build(self):
        r"""Populates the toolbar with items."""
        # This function is called first when no experiment has been loaded yet.
        try:
            self.experiment
        except:
            return
        self.clear()
        if self.orientation() == QtCore.Qt.Vertical:
            self.addWidget(ToolbarItemsLabel(self, _(u'Commonly used')))
        # Add the core items
        self.add_content([ToolbarItemsItem(self, item)
                          for item in self.experiment.core_items])
        # Create a dictionary of plugins by category. We also maintain a list
        # to preserve the order of the categories.
        cat_dict = OrderedDict()
        for plugin in self.plugin_manager.filter(modes=self.main_window.mode):
            # Ignoring disabled extensions
            cfg_key = f'plugin_enabled_{plugin.name}'
            if cfg_key in cfg and not cfg[cfg_key]:
                continue
            cat = plugin['category']
            if cat not in cat_dict:
                cat_dict[cat] = []
            cat_dict[cat].append(plugin)
        # Add the plugins
        for cat, cat_plugins in cat_dict.items():
            self.addSeparator()
            if self.orientation() == QtCore.Qt.Vertical:
                self.addWidget(ToolbarItemsLabel(self, cat))
            content = []
            for plugin in cat_plugins:
                oslogger.debug(u"adding plugin '%s'" % plugin)
                pixmap = self.theme.qpixmap(plugin.icon_large)
                content.append(ToolbarItemsItem(self, plugin.name, pixmap))
            self.add_content(content)

    def collapse(self):
        r"""Collapses the item toolbar if is was expanded."""
        if self._expand_button.isChecked():
            self._expand_button.click()
            
    def _refresh_enabled(self):
        for item in self._items:
            item.setEnabled(self.experiment.items.is_supported(item.item))

    def event_change_experiment(self): self._refresh_enabled()
    def event_open_experiment(self, path): self._refresh_enabled()
    def event_startup(self): self._refresh_enabled()


# Alias for backwards compatibility
toolbar_items = ToolbarItems



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtCore, QtWidgets
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'tree_base_item', category=u'core')


class TreeBaseItem(BaseSubcomponent, QtWidgets.QTreeWidgetItem):

    r"""A base class that corresponds to any widget in the overview area."""
    def start_edit(self, col):

        self.treeWidget().editItem(self, col)

    def ancestry(self):

        return None

    def path(self):

        if self.parent() is None:
            return self.text(0)
        return self.parent().path() + u'.' + self.text(0)

    def children(self):

        l = []
        for i in range(self.childCount()):
            child = self.child(i)
            l += [child] + child.children()
        return l

    def show_context_menu(self, pos):

        pass

    def paste(self):

        pass

    def copy(self):

        pass

    def copy_linked(self):

        pass

    def copy_unlinked(self):

        pass
    
    def copy_shallow(self):
        
        pass

    def delete(self):

        pass

    def permanently_delete(self):

        pass

    def open_tab(self):

        pass

    def drop_hint(self):

        return None

    def droppable(self, data):

        return self._droppable

    def set_droppable(self, droppable):

        self._droppabe = droppable

    def draggable(self):

        return self._draggable

    def set_draggable(self, draggable):

        self._draggable = draggable

    def has_append_menu(self):

        return False

    def expand(self):
        r"""Expands this item and all items under it."""
        self.setExpanded(True)
        for i in range(self.childCount()):
            self.child(i).expand()

    def collapse(self):
        r"""Collapses this item and all items under it."""
        self.setExpanded(False)
        for i in range(self.childCount()):
            self.child(i).collapse()

    def rename(self, from_name, to_name):
        r"""Renames an item.

        Parameters
        ----------
        from_name : unicode
            The old name.
        to_name : unicode
            The new name.
        """
        for i in range(self.childCount()):
            self.child(i).rename(from_name, to_name)

    def set_icon(self, name, icon):
        r"""Changes an item's icon.

        Parameters
        ----------
        name : unicode
            The item name.
        icon : unicode
            The icon name.
        """
        for i in range(self.childCount()):
            self.child(i).set_icon(name, icon)


# Alias for backwards compatibility
tree_base_item = TreeBaseItem



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame import sketchpad_elements
from libopensesame.item import Item
from libopensesame.exceptions import InvalidSketchpadElementScript, \
    InvalidValue
from libopensesame.base_response_item import BaseResponseItem
from libopensesame.keyboard_response import KeyboardResponseMixin
from libopensesame.mouse_response import MouseResponseMixin
from openexp.canvas import Canvas


class Sketchpad(BaseResponseItem, KeyboardResponseMixin, MouseResponseMixin):

    description = 'Displays stimuli'
    is_oneshot_coroutine = True

    def reset(self):
        self.var.duration = 'keypress'
        self.elements = []

    def element_module(self):
        r"""Determines the module to be used for the element classes. The
        runtime and GUI use different modules.

        Returns
        -------
        module
            A module containing sketchpad-element classes
        """
        return sketchpad_elements

    def from_string(self, string):
        self.var.clear()
        self.comments = []
        self.reset()
        if string is None:
            return
        for line in string.split('\n'):
            if self.parse_variable(line):
                continue
            cmd, arglist, kwdict = self.syntax.parse_cmd(line)
            if cmd != 'draw':
                continue
            if len(arglist) == 0:
                raise InvalidSketchpadElementScript(
                    f'Incomplete draw command: {line}')
            element_type = arglist[0]
            if not hasattr(self.element_module(), element_type):
                raise InvalidSketchpadElementScript(
                    f'Unknown sketchpad element: {element_type}')
            element_class = getattr(self.element_module(), element_type)
            element = element_class(self, line)
            self.elements.append(element)
        self.elements.sort(
            key=lambda element:
            -element.z_index
            if isinstance(element.z_index, int)
            else 0
        )

    def process_response(self, response_args):
        """See base_response_item."""
        if self.var.duration == 'mouseclick':
            MouseResponseMixin.process_response(self, response_args)
            return
        super().process_response(response_args)

    def prepare_response_func(self):
        """See base_response_item."""
        if isinstance(self.var.duration, (int, float)):
            self._flush = lambda: None
            return self._prepare_sleep_func(self.var.duration)
        if self.var.duration == 'keypress':
            self._flush = lambda: self._keyboard.flush()
            return KeyboardResponseMixin.prepare_response_func(self)
        if self.var.duration == 'mouseclick':
            self._flush = lambda: self._mouse.flush()
            return MouseResponseMixin.prepare_response_func(self)
        raise InvalidValue(f'Invalid duration: {self.var.duration}')

    def _elements(self):
        r"""Creates a list of sketchpad elements that are shown, sorted by
        z-index.
        """
        elements = [e for e in self.elements if e.is_shown()]
        try:
            elements.sort(key=lambda e: -int(self.syntax.eval_text(e.z_index)))
        except ValueError as e:
            raise InvalidValuei('Invalid z_index for sketchpad element')
        return elements

    def prepare(self):
        super().prepare()
        self.canvas = Canvas(self.experiment, color=self.var.foreground,
                             background_color=self.var.background)
        with self.canvas:
            for element in self._elements():
                temp_name = element.draw()
                if element.element_name is not None:
                    self.canvas.rename_element(temp_name, element.element_name)

    def run(self):
        self._t0 = self.set_item_onset(self.canvas.show())
        self._flush()
        super().run()

    def coroutine(self):
        yield
        self.set_item_onset(self.canvas.show())

    def to_string(self):
        s = super().to_string()
        for element in self.elements:
            s += '\t%s\n' % element.to_string()
        return s

    def var_info(self):
        if self.var.get('duration', _eval=False, default='') in \
                ['keypress', 'mouseclick']:
            return super().var_info()
        return Item.var_info(self)


# Alias for backwards compatibility
sketchpad = Sketchpad



"""Turns a tab into a dockwidget"""

icon = "view-fullscreen"
label = "Tab to dockwidget"
priority = -1
menu = {
    "index": 3,
    "separator_after": False,
    "separator_before": False,
    "submenu": "View"
}
toolbar = {
    "index": 9,
    "separator_after": False,
    "separator_before": False
}
tooltip = "Turn tab into dockwidget"



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import operator


class CIStr(str):

    r"""A class for case-insensitive string comparisons. This is used to speed
    up the item store, which gets a lot of look ups for item names. This is not
    a full emulation of the str protocol, only those parts that are used by the
    item store.
    """
    def __init__(self, s):

        str.__init__(s)
        self._lower = s.lower()
        self._hash = hash(self._lower)

    def __eq__(self, other):

        try:
            return other._hash == self._hash
        except AttributeError:
            if hasattr(other, u'lower'):
                return other.lower() == self._lower
        return other == self._lower

    def __ne__(self, other):

        try:
            return other._hash != self._hash
        except AttributeError:
            if hasattr(other, u'lower'):
                return other.lower() != self._lower
        return other != self._lower

    def __hash__(self):

        return self._hash


# Alias for backwards compatibility
cistr = CIStr



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qdatamatrix import QDataMatrix
from pseudorandom import EnforceFailed
from libopensesame.loop import Loop as LoopRuntime
from libopensesame.exceptions import OSException
from libqtopensesame.items.qtitem import QtItem
from libqtopensesame.items.qtstructure_item import QtStructureItem
from libqtopensesame.widgets.loop_widget import LoopWidget
from libqtopensesame.widgets.tree_item_item import TreeItemItem
from libqtopensesame.misc.translate import translation_context
import math
_ = translation_context(u'loop', category=u'item')


class Loop(QtStructureItem, QtItem, LoopRuntime):

    """The GUI for the loop item"""
    description = _(u'Repeatedly runs another item')
    help_url = u'manual/structure/loop'
    lazy_init = True

    def __init__(self, name, experiment, string=None):
        """See QtItem."""
        self.lock_cycles = False
        LoopRuntime.__init__(self, name, experiment, string)
        QtStructureItem.__init__(self)
        QtItem.__init__(self)

    def init_edit_widget(self):
        """Builds the loop controls."""
        super().init_edit_widget(stretch=False)
        self.loop_widget = LoopWidget(self.experiment.main_window)
        self.qdm = QDataMatrix(self.dm)
        self.qdm.changed.connect(self._apply_table)
        self.qdm.cellchanged.connect(self._on_cell_changed)
        self.edit_vbox.addWidget(self.loop_widget)
        self.edit_vbox.addWidget(self.qdm)
        self.set_focus_widget(self.qdm)
        self.auto_add_widget(self.loop_widget.ui.spinbox_repeat, u'repeat')
        self.auto_add_widget(self.loop_widget.ui.combobox_order, u'order')
        self.auto_add_widget(self.loop_widget.ui.combobox_source, u'source',
                             apply_func=self._apply_source)
        self.auto_add_widget(
            self.loop_widget.ui.edit_source_file, u'source_file')
        self.loop_widget.ui.edit_break_if.verb = 'break'
        self.auto_add_widget(self.loop_widget.ui.edit_break_if, u'break_if')
        self.auto_add_widget(self.loop_widget.ui.checkbox_break_if_on_first,
                             u'break_if_on_first')
        self.auto_add_widget(self.loop_widget.ui.checkbox_continuous,
                             u'continuous')
        self.loop_widget.ui.combobox_item.activated.connect(self._apply_item)
        self.loop_widget.ui.button_preview.clicked.connect(self._show_preview)
        self.loop_widget.ui.button_wizard.clicked.connect(self._show_wizard)

    def _show_wizard(self):
        r"""Shows the full-factorial-design wizard."""
        from libqtopensesame.dialogs.loop_wizard import loop_wizard
        d = loop_wizard(self.main_window)
        dm = d.exec_()
        if dm is None:
            return
        dm.sorted = False
        self.dm = dm
        self.update()

    def _show_preview(self):
        r"""Shows a preview tab for the loop table."""
        l = [_(u'# Preview of loop table'), u'\n\n']
        try:
            dm = self._create_live_datamatrix()
        except (OSException, EnforceFailed) as e:
            l.append(_(u'Failed to generate preview.'))
            l.append(u'\n~~~ .python\n%s\n~~~' % e)
        else:
            l.append(u'\n\n<table><thead><tr>')
            l.append(u''.join([u'<th>%s</th>' % column_name
                               for column_name in dm.column_names]))
            l.append(u'</tr></thead><tbody>')
            for row in dm:
                l.append(u'<tr>'
                         + u''.join([u'<th>%s</th>' %
                                    val for name, val in row])
                         + u'</tr>')
            l.append(u'</tbody></table>')
        md = u'\n'.join(l)
        self.tabwidget.open_markdown(
            md, title=u'Loop preview', icon=u'os-loop')

    @QtStructureItem.clears_children_cache
    def _apply_item(self, *args):
        r"""Applies changes to the item combobox.

        Parameters
        ----------
        *args : list
            Dummy arguments passed by the signal-slot system.
        """
        self._item = self.loop_widget.ui.combobox_item.selected_item
        self.experiment.build_item_tree()
        self.update_script()

    def _update_item(self):
        r"""Updates the item combobox."""
        self.loop_widget.ui.combobox_item.select(self._item)

    def _on_cell_changed(self, rownr, colnr):
        r"""Protects the table from having newlines in the cells."""
        if rownr == 0:
            return
        col = self.qdm.dm.columns[colnr][1]
        val = self.qdm.dm[col][rownr - 1]
        if not isinstance(val, str) or u'\n' not in val:
            return
        val = val.replace(u'\n', u'')
        self.qdm._spreadsheet._setcell(rownr, colnr, val)
        self.qdm.dm[col][rownr - 1] = val
        self._apply_table()

    def _apply_table(self):
        r"""Applies changes to the loop table."""
        # QDataMatrix can create a new datamatrix object, and that's the one we
        # have to use.
        self.dm = self.qdm.dm
        if not self.dm:
            self.dm.length = 1
            self.qdm.refresh()
        if not self.dm.columns:
            self.dm.empty_column = u''
            self.qdm.refresh()
        self.update_script()
        self._warn_empty_rows()

    def _apply_source(self):
        r"""Applies changes to the source selector, if a file source is
        selected.
        """
        self.var.source = self.loop_widget.ui.combobox_source.currentText()
        self._update_source()
        self.update_script()

    def _update_source(self):
        r"""Update the file-source selector, hiding it when the loop table is
        used as source.
        """
        file_mode = self.var.get(u'source', _eval=False) != u'table'
        self.loop_widget.ui.label_source_file.setVisible(file_mode)
        self.loop_widget.ui.edit_source_file.setVisible(file_mode)
        self.loop_widget.ui.spacer.setVisible(file_mode)
        self.qdm.setVisible(not file_mode)

    def _row_count_text(self, n):
        r"""A descriptive text for the number of rows.

        Parameters
        ----------
        n
            The number of rows.

        Returns
        -------
        A descriptive text.
        """
        if n == 1:
            return _(u'one row occurs')
        if n == 2:
            return _(u'two rows occur')
        return _(u'%s rows occur' % n)

    def _time_count_text(self, n):
        r"""A descriptive text for the number of times an item is executed.

        Parameters
        ----------
        n
            The number of times.

        Returns
        -------
        A descriptive text.
        """
        if n == 0:
            return _(u'never')
        if n == 1:
            return _(u'once')
        if n == 2:
            return _(u'twice')
        return _(u'%s times' % n)

    def _set_summary(self, msg):
        r"""Sets the summary label.

        Parameters
        ----------
        msg
            The summary message.
        """
        self.loop_widget.ui.label_summary.setText(msg)

    def _update_summary(self):
        r"""Generates a loop summary and sets the summary label."""
        if self._item not in self.experiment.items:
            self._set_summary(_(u'Warning: No item to run has been specified'))
            return
        repeat = self.var.get(u'repeat', _eval=False)
        order = self.var.get(u'order', _eval=False)
        if not isinstance(repeat, (float, int)) \
                or order not in self.valid_orders:
            self._set_summary(
                _(u'Note: Order or repeat is unknown or variably defined'))
            return
        if self.var.get(u'source', _eval=False) != u'table':
            self._set_summary(
                _(u'Note: Loop data will be read from file'))
            return

        f, i = math.modf(repeat)
        length = int(repeat*len(self.dm))
        numrows = len(self.dm)
        r_freq = int(i+1)
        n_freq = int(f * numrows)
        r_rare = int(i)
        n_rare = numrows - n_freq
        s = _(u'Summary: <b>%s</b> will be called <b>%s</b> in <b>%s</b> order.') \
            % (self._item, self._time_count_text(length), order)
        s += u' ' + _(u'The number of rows is %s.') % numrows
        if n_freq == 0:
            s += u' ' + \
                _(u'All rows occur %s.') % self._time_count_text(r_rare)
        elif n_rare == 0:
            s += u' ' + \
                _(u'All rows occur %s.') % self._time_count_text(r_freq)
        else:
            s += u' ' + _(u'This means that %s %s and %s %s.') \
                % (self._row_count_text(n_freq),
                   self._time_count_text(r_freq),
                   self._row_count_text(n_rare),
                   self._time_count_text(r_rare))
        self._set_summary(s)

    def _warn_empty_rows(self):
        r"""Gives a notification when the loop table has more than 1 rows and
        the last row is empty. This generally happens if the user has forgotten
        to truncate the loop table when clearing rows at the end.
        """
        if len(self.qdm.dm) > 1 and all(
                cell == u'' for name, cell in self.qdm.dm[-1]
        ):
            self.extension_manager.fire(
                u'notify',
                message=u'The loop table has empty rows at the end'
            )

    def edit_widget(self):
        """See QtItem."""
        super().edit_widget()
        self.loop_widget.ui.combobox_item.filter_fnc = (
            lambda item: item not in self.parents()
        )
        self.loop_widget.ui.combobox_item.refresh()
        self.loop_widget.ui.combobox_item.select(self._item)
        self.qdm.dm = self.dm
        self.qdm.refresh()
        self._update_summary()
        self._update_item()
        self._update_source()
        self._warn_empty_rows()

    def update_script(self, use_cache=True):
        """See QtItem."""
        super().update_script(use_cache=use_cache)
        self._update_summary()

    def build_item_tree(self, toplevel=None, items=[], max_depth=-1,
                        extra_info=None):
        """See QtItem."""
        items.append(self.name)
        widget = TreeItemItem(self, extra_info=extra_info)
        if toplevel is not None:
            toplevel.addChild(widget)
        if (max_depth < 0 or max_depth > 1) \
                and self._item in self.experiment.items:
            self.experiment.items[self._item].build_item_tree(widget, items,
                max_depth=max_depth-1)
        return widget

    @QtStructureItem.cached_children
    def children(self):
        """See QtItem."""
        if self._item not in self.experiment.items:
            return []
        return [self._item] + self.experiment.items[self._item].children()

    def direct_children(self):

        return [self._item] if self._item in self.experiment.items else []

    def is_child_item(self, item):
        """See QtItem."""
        return self._item == item or (self._item in self.experiment.items and
                                      self.experiment.items[self._item].is_child_item(item))

    @QtStructureItem.clears_children_cache
    def insert_child_item(self, item_name, index=0):
        """See QtItem."""
        self._item = item_name
        self.update()
        self.main_window.set_unsaved(True)

    @QtStructureItem.clears_children_cache
    def remove_child_item(self, item_name, index=0):
        """See QtItem."""
        if item_name == self._item:
            self._item = u''
        if not self.update():
            self.extension_manager.fire(u'change_item', name=self.name)
        self.main_window.set_unsaved(True)

    @QtStructureItem.clears_children_cache
    def rename(self, from_name, to_name):
        """See QtItem."""
        super().rename(from_name, to_name)
        if self._item == from_name:
            self._item = to_name

    @QtStructureItem.clears_children_cache
    def delete(self, item_name, item_parent=None, index=None):
        """See QtItem."""
        if self._item == item_name and item_parent == self.name:
            self._item = u''


# Alias for backwards compatibility
loop = Loop



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.sampler import Sampler as SamplerRuntime
from libqtopensesame.items.qtplugin import QtPlugin
from libqtopensesame.validators import DurationValidator
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'sampler', category=u'item')


class Sampler(SamplerRuntime, QtPlugin):

    r"""GUI controls for the sampler item."""
    description = _(u'Plays a sound file in .wav or .ogg format')
    help_url = u'manual/stimuli/sound'
    lazy_init = True

    def __init__(self, name, experiment, string=None):
        """See item."""
        SamplerRuntime.__init__(self, name, experiment, string)
        QtPlugin.__init__(self)

    def init_edit_widget(self):
        """See qtitem."""
        QtPlugin.init_edit_widget(self)
        self.add_filepool_control(u'sample', _(u'Sound file'),
                                  info=_(u'In .ogg or .wav format'))
        self.add_doublespinbox_control(u'volume', _(u'Volume'),
                                       min_val=0, max_val=1,
                                       suffix=_(u' x original'))
        self.add_line_edit_control(u'pan', _(u'Panning'),
                                   info=_(u'Positive values toward the right; "left" or "right" for full panning'))
        self.add_doublespinbox_control(u'pitch', _(u'Pitch'), min_val=0,
                                       max_val=1000, suffix=_(u' x original'))
        self.add_spinbox_control(u'stop_after', _(u'Stop after'),
                                 min_val=0, max_val=10000000, suffix=_(u' ms'))
        self.add_spinbox_control(u'fade_in', _(u'Fade in'),
                                 min_val=0, max_val=10000000, suffix=_(u' ms'))
        self.add_line_edit_control(u'duration', _(u'Duration'),
                                   info=_(u'In milliseconds, "sound", "keypress", or "mouseclick"'),
                                   validator=DurationValidator(self, default=u'sound'))


# Alias for backwards compatibilit
sampler = Sampler



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.exceptions import OSException
from libopensesame.base_response_item import BaseResponseItem
from libopensesame.sampler import Sampler
from openexp.synth import Synth as OpenExpSynth


class Synth(Sampler):
    """An item for synthesized-sound playback."""
    
    description = 'A basic sound synthesizer'

    def reset(self):
        self.var.freq = 440
        self.var.length = 100
        self.var.osc = 'sine'
        self.var.pan = 0
        self.var.attack = 0
        self.var.decay = 5
        self.var.volume = 1.0
        self.var.duration = 'sound'
        self.block = False

    def prepare(self):
        BaseResponseItem.prepare(self)
        try:
            self.sampler = OpenExpSynth(self.experiment, osc=self.var.osc,
                                        freq=self.var.freq,
                                        length=self.var.length,
                                        attack=self.var.attack,
                                        decay=self.var.decay)
        except OSException as e:
            raise
        except Exception as e:
            raise OSException(f'Failed to generate sound: {e}')
        pan = self.var.pan
        if pan == -20:
            pan = 'left'
        elif pan == 20:
            pan = 'right'
        self.sampler.pan = pan
        self.sampler.volume = self.var.volume
        self.sampler.block = self.var.duration == 'sound'


# Alias for backwards compatibility
synth = Synth



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import pygame
from openexp._canvas import canvas
from openexp._canvas._noise_patch.noise_patch import NoisePatch
from openexp._canvas._element.xpyriment import XpyrimentElement
from expyriment.stimuli._visual import Visual


class Xpyriment(XpyrimentElement, NoisePatch):

    def prepare(self):

        im = canvas._noise_patch(
            self.env,
            self.size,
            self.stdev,
            self.col1,
            self.col2,
            self.bgmode
        )
        surface = pygame.image.fromstring(im.tobytes(), im.size, u'RGB')
        self._stim = Visual(position=self.to_xy(self.x, self.y))
        self._stim.set_surface(surface)
        self._stim.preload()



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.exceptions import OSException
from ._base_task import BaseTask
import inspect


class InlineTask(BaseTask):

    r"""A task controls the coroutine for a Python generator function."""
    def __init__(self, coroutines, function_name, python_workspace, start_time,
                 end_time):
        r"""Constructor.

        Parameters
        ----------
        function_name : str
            The name of a Python generator function.
        python_workspace : python_workspace
            The python-workspace object.
        """
        super().__init__(coroutines, start_time, end_time)
        self.function_name = function_name
        self.python_workspace = python_workspace

    def launch(self):
        """See base_task."""
        if self.function_name not in self.python_workspace._globals:
            raise OSException(u'"%s" is not defined' % self.function_name)
        self.coroutine = self.python_workspace._globals[self.function_name]
        if not inspect.isgeneratorfunction(self.coroutine):
            raise OSException(
                u'"%s" is not a generator function' % self.function_name)
        self.coroutines.event('launching %s' % self.coroutine)
        self.coroutine = self.coroutine()
        self.coroutine.send(None)



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtWidgets
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'confirmation', category=u'core')


class Confirmation(QtWidgets.QMessageBox, BaseSubcomponent):

    r"""A simple yes/ no/ cancel confirmation dialog."""
    def __init__(self, main_window, msg, title=None, allow_cancel=False,
                 default=u'no'):
        r"""Constructor.

        Parameters
        ----------
        main_window : QWidget
            The main-window object.
        msg : unicode, str
            The message.
        title : str, NoneType
            A Window title or None for a default title.
        allow_cancel : bool
            Indicates whether a cancel button should be included, in addition
            to the yes and no buttons.
        default : str
            The button that is active by default, 'no', 'yes', or 'cancel'
        """
        QtWidgets.QMessageBox.__init__(self, main_window)
        self.setup(main_window)
        self.yes = self.addButton(QtWidgets.QMessageBox.Yes)
        self.no = self.addButton(QtWidgets.QMessageBox.No)
        if allow_cancel:
            self.cancel = self.addButton(QtWidgets.QMessageBox.Cancel)
        else:
            self.cancel = None
        if default == u'no':
            self.setDefaultButton(QtWidgets.QMessageBox.No)
        elif default == u'yes':
            self.setDefaultButton(QtWidgets.QMessageBox.Yes)
        elif default == u'cancel' and allow_cancel:
            self.setDefaultButton(QtWidgets.QMessageBox.Cancel)
        if title is None:
            title = _(u'Please confirm')
        self.setWindowTitle(title)
        self.setText(msg)

    def show(self):
        r"""Shows the confirmation dialog.

        Returns
        -------
        bool, NoneType
            True if confirmed, False disconfirmed, and None if cancelled.
        """
        self.exec_()
        button = self.clickedButton()
        if self.cancel is not None and button is self.cancel:
            return None
        return button == self.yes


# Alias for backwards compatibility
confirmation = Confirmation



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
from libqtopensesame.widgets.tree_append_menu import TreeAppendMenu
from qtpy import QtWidgets


class TreeAppendButton(BaseSubcomponent, QtWidgets.QPushButton):

    r"""An append item menu that appears below the tree structure of sequence
    items.
    """
    def setup(self, main_window):

        super().setup(main_window)
        self.setIcon(self.main_window.theme.qicon(u'list-add'))
        self.tree_overview = self.parent()
        self.append_menu = TreeAppendMenu(self.tree_overview)
        self.setMenu(self.append_menu)
        self.setFlat(True)
        self.adjustSize()

    def set_position(self):
        r"""Places the button right below the last item."""
        last_item = self.tree_overview.topLevelItem(0)
        while self.tree_overview.itemBelow(last_item) is not None:
            last_item = self.tree_overview.itemBelow(last_item)
        index = self.tree_overview.indexFromItem(last_item)
        rect = self.tree_overview.visualRect(index)
        rect.moveTop(int(rect.top() + 1.75*rect.height()))
        rect.moveLeft(4)
        geom = self.geometry()
        geom.moveTopLeft(rect.bottomLeft())
        self.setGeometry(geom)


# Alias for backwards compatibility
tree_append_button = TreeAppendButton
